import os
import socket
from libensemble.resources.env_resources import EnvResources
from libensemble.resources.resources import Resources, GlobalResources, ResourcesException
from libensemble.resources.worker_resources import ResourceManager, WorkerResources
from libensemble.resources.mpi_resources import create_machinefile


def setup_standalone_run():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    if os.path.isfile('node_list'):
        os.remove('node_list')


def teardown_standalone_run():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    if os.path.isfile('node_list'):
        os.remove('node_list')


def setup_function(function):
    print("setup_function    function:%s" % function.__name__)
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    # if os.environ['LIBE_RESOURCES_TEST_NODE_LIST']:
    #     del os.environ['LIBE_RESOURCES_TEST_NODE_LIST']
    # if os.environ['THIS_ENV_VARIABLE_IS_DEF_NOT_SET']:
    #     del os.environ['THIS_ENV_VARIABLE_IS_DEF_NOT_SET']
    if os.path.isfile('node_list'):
        os.remove('node_list')


def teardown_function(function):
    print("teardown_function    function:%s" % function.__name__)
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    # if os.environ['LIBE_RESOURCES_TEST_NODE_LIST']:
    #     del os.environ['LIBE_RESOURCES_TEST_NODE_LIST']
    # if os.environ['THIS_ENV_VARIABLE_IS_DEF_NOT_SET']:
    #     del os.environ['THIS_ENV_VARIABLE_IS_DEF_NOT_SET']
    if os.path.isfile('node_list'):
        os.remove('node_list')


def sname(name):
    print('sname being set')
    return name.split(".", 1)[0]


# Tests ========================================================================================

# Tests GlobalResources.get_global_nodelist (This requires above tests to work)
def test_get_global_nodelist_frm_slurm():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0137-0139,1234]"
    exp_out = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_slurm_suffix():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0137-0139,1234].eps"
    exp_out = ['knl-0020.eps', 'knl-0021.eps', 'knl-0022.eps', 'knl-0137.eps',
               'knl-0138.eps', 'knl-0139.eps', 'knl-1234.eps']
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_slurm_single():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-0020"
    exp_out = ['knl-0020']
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_slurm_straight_list():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "n0000.es1,n0002.es1"
    exp_out = ['n0000.es1', 'n0002.es1']
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


# Default also sorts
def test_get_global_nodelist_frm_slurm_multigroup():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0137-0139,1234],bdw[021,455]ext"
    exp_out = ['bdw021ext', 'bdw455ext', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_slurm_complex():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "n0000.es1,knl-[0137-0139,1234],n0002.es1,bds[021,455]-ext"
    exp_out = ['bds021-ext', 'bds455-ext', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234', 'n0000.es1', 'n0002.es1']
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_cobalt():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "20-22,137-139,1234"
    exp_out = ['20', '21', '22', '137', '138', '139', '1234']
    env_resources = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_cobalt="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_lsf():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5' + ' g06n02' * 42 + ' h21n18' * 42
    exp_out = ['g06n02', 'h21n18']
    env_resources = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_frm_lsf_shortform():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5 1 g06n02 42 h21n18 42'
    exp_out = ['g06n02', 'h21n18']
    env_resources = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="LIBE_RESOURCES_TEST_NODE_LIST")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == exp_out, \
        "global_nodelist returned does not match expected. \nRet: {}\nExp: {}".format(global_nodelist, exp_out)


def test_get_global_nodelist_standalone():
    mynode = socket.gethostname()
    exp_node = mynode  # sname(mynode)
    env_resources = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist == [exp_node], "global_nodelist returned does not match expected"


def test_get_global_nodelist_frm_wrklst_file():
    # node_list file should override env variables
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "20-22,137-139,1234"  # Should not be this
    exp_out = ['knl-0019', 'knl-0021', 'knl-0022', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-2345']  # Should be this

    open('node_list', 'w').close()
    try:
        _ = GlobalResources.get_global_nodelist(rundir=os.getcwd())
    except ResourcesException as e:
        assert e.args[0] == 'Error. global_nodelist is empty'
    else:
        assert 0

    with open('node_list', 'w') as f:
        for node in exp_out:
            f.write(node + '\n')

    # Do not specify env vars.
    global_nodelist1 = GlobalResources.get_global_nodelist(rundir=os.getcwd())
    assert global_nodelist1 == exp_out, "global_nodelist returned does not match expected"

    # Specify env vars - should ignore
    env_resources = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_cobalt="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    global_nodelist2 = GlobalResources.get_global_nodelist(rundir=os.getcwd(), env_resources=env_resources)
    assert global_nodelist2 == exp_out, "global_nodelist returned does not match expected"
    os.remove('node_list')


def test_remove_libE_nodes():
    mynode = socket.gethostname()
    exp_out = ['knl-0019', 'knl-0021', 'knl-0022', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-2345']

    # Add at beginning
    nodes_in = [mynode] + exp_out
    nodes_out = GlobalResources.remove_nodes(nodes_in, mynode)
    assert nodes_out == exp_out, "nodelist returned does not match expected"

    # Add twice in middle and at end
    nodes_in = []
    for i, node in enumerate(exp_out):
        nodes_in.append(node)
        if i == 1 or i == 4 or i == 6:
            nodes_in.append(mynode)
    nodes_out = GlobalResources.remove_nodes(nodes_in, mynode)
    assert nodes_out == exp_out, "nodelist returned does not match expected"


def _assert_worker_attr(wres, attr, exp):
    ret = getattr(wres, attr)
    assert ret == exp, "{} returned does not match expected.  \nRet: {}\nExp: {}".format(attr, ret, exp)


# These are all 1 worker per rset.
def _worker_asserts(wres, split_list, exp_slots, wrk, nworkers, nnodes, reps=1):

    # Create dictionary of attributes and expected values
    exp_dict = {'workerID': wrk + 1,
                'local_nodelist': split_list[wrk],
                'slots': exp_slots,
                'local_node_count': nnodes,
                'num_workers': nworkers,
                'split_list': split_list,
                'slot_count': 1,
                'local_rsets_list': [reps] * nworkers,
                'rsets_per_node': reps,
                }

    for attr, exp_val in exp_dict.items():
        _assert_worker_attr(wres, attr, exp_val)


# These are all >= 1 node per rset. And 1 worker per rset
# dedicated_mode makes no difference in this test
def test_get_local_resources_dedicated_mode():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0036,0137-0139,1234]"
    resource_info = {'nodelist_env_slurm': "LIBE_RESOURCES_TEST_NODE_LIST"}
    libE_specs = {'resource_info': resource_info,
                  'dedicated_mode': True}
    gresources = GlobalResources(libE_specs)

    # 8 Workers ---------------------------------------------------------------
    nworkers = 8
    exp_out = [['knl-0020'], ['knl-0021'], ['knl-0022'], ['knl-0036'],
               ['knl-0137'], ['knl-0138'], ['knl-0139'], ['knl-1234']]

    # Spoof current process as each worker and check nodelist and other worker resources.
    for wrk in range(nworkers):
        workerID = wrk + 1
        exp_slots = {exp_out[wrk][0]: [0]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 1)
        del wresources

    # 4 Workers ---------------------------------------------------------------
    nworkers = 4
    exp_out = [['knl-0020', 'knl-0021'], ['knl-0022', 'knl-0036'],
               ['knl-0137', 'knl-0138'], ['knl-0139', 'knl-1234']]
    for wrk in range(nworkers):
        workerID = wrk + 1
        exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 2)
        del wresources

    # 1 Worker ----------------------------------------------------------------
    nworkers = 1
    exp_out = [['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036',
                'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']]

    # Just write out this one as one worker.
    exp_slots = {'knl-0020': [0], 'knl-0021': [0], 'knl-0022': [0],
                 'knl-0036': [0], 'knl-0137': [0], 'knl-0138': [0],
                 'knl-0139': [0], 'knl-1234': [0]}

    for wrk in range(nworkers):
        workerID = wrk + 1
        # exp_slots - see above
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 8)
        del wresources

    # 3 Workers (Test the best_split algorithm) -------------------------------
    nworkers = 3
    exp_out = [['knl-0020', 'knl-0021', 'knl-0022'],
               ['knl-0036', 'knl-0137', 'knl-0138'],
               ['knl-0139', 'knl-1234']]
    for wrk in range(nworkers):
        workerID = wrk + 1
        exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])

        # Trying to avoid generic algorithm
        if len(wresources.local_nodelist) == 3:
            exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0], exp_out[wrk][2]: [0]}
            nnodes = 3
        else:
            exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0]}
            nnodes = 2
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, nnodes)

    # 16 Workers --------------------------------------------------------------
    # Multiple workers per node
    nworkers = 16

    exp_out = [['knl-0020'], ['knl-0020'], ['knl-0021'], ['knl-0021'],
               ['knl-0022'], ['knl-0022'], ['knl-0036'], ['knl-0036'],
               ['knl-0137'], ['knl-0137'], ['knl-0138'], ['knl-0138'],
               ['knl-0139'], ['knl-0139'], ['knl-1234'], ['knl-1234']]

    for wrk in range(nworkers):
        workerID = wrk + 1

        # If even worker have slot 1
        if (workerID % 2) == 0:
            myslot = 1
        else:
            myslot = 0
        exp_slots = {exp_out[wrk][0]: [myslot]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 1, 2)

    del os.environ["LIBE_RESOURCES_TEST_NODE_LIST"]


# The main tests are same as above - note for when fixtures set up
def test_get_local_resources_dedicated_mode_remove_libE_proc():
    mynode = socket.gethostname()
    nodelist_in = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    with open('node_list', 'w') as f:
        for i, node in enumerate(nodelist_in):
            f.write(node + '\n')
            if i == 3:
                f.write(mynode + '\n')

    libE_specs = {'dedicated_mode': True}

    gresources = GlobalResources(libE_specs)
    gresources.add_comm_info(libE_nodes=[mynode])

    # 8 Workers ---------------------------------------------------------------
    nworkers = 8
    exp_out = [['knl-0020'], ['knl-0021'], ['knl-0022'], ['knl-0036'],
               ['knl-0137'], ['knl-0138'], ['knl-0139'], ['knl-1234']]

    # Spoof current process as each worker and check nodelist and other worker resources.
    for wrk in range(nworkers):
        workerID = wrk + 1
        exp_slots = {exp_out[wrk][0]: [0]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 1)
        del wresources

    # 4 Workers ---------------------------------------------------------------
    nworkers = 4
    exp_out = [['knl-0020', 'knl-0021'], ['knl-0022', 'knl-0036'],
               ['knl-0137', 'knl-0138'], ['knl-0139', 'knl-1234']]
    for wrk in range(nworkers):
        workerID = wrk + 1
        exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 2)
        del wresources

    # 1 Worker ----------------------------------------------------------------
    nworkers = 1
    exp_out = [['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036',
                'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']]

    # Just write out this one as one worker.
    exp_slots = {'knl-0020': [0], 'knl-0021': [0], 'knl-0022': [0],
                 'knl-0036': [0], 'knl-0137': [0], 'knl-0138': [0],
                 'knl-0139': [0], 'knl-1234': [0]}

    for wrk in range(nworkers):
        workerID = wrk + 1
        # exp_slots - see above
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 8)
        del wresources

    # 3 Workers (Test the best_split algorithm) -------------------------------
    nworkers = 3
    exp_out = [['knl-0020', 'knl-0021', 'knl-0022'],
               ['knl-0036', 'knl-0137', 'knl-0138'],
               ['knl-0139', 'knl-1234']]
    for wrk in range(nworkers):
        workerID = wrk + 1
        exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])

        # Trying to avoid generic algorithm
        if len(wresources.local_nodelist) == 3:
            exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0], exp_out[wrk][2]: [0]}
            nnodes = 3
        else:
            exp_slots = {exp_out[wrk][0]: [0], exp_out[wrk][1]: [0]}
            nnodes = 2
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, nnodes)
        del wresources

    # 16 Workers --------------------------------------------------------------
    # Multiple workers per node
    nworkers = 16

    exp_out = [['knl-0020'], ['knl-0020'], ['knl-0021'], ['knl-0021'],
               ['knl-0022'], ['knl-0022'], ['knl-0036'], ['knl-0036'],
               ['knl-0137'], ['knl-0137'], ['knl-0138'], ['knl-0138'],
               ['knl-0139'], ['knl-0139'], ['knl-1234'], ['knl-1234']]

    for wrk in range(nworkers):
        workerID = wrk + 1

        # If even worker have slot 1
        if (workerID % 2) == 0:
            myslot = 1
        else:
            myslot = 0
        exp_slots = {exp_out[wrk][0]: [myslot]}
        wresources = WorkerResources(nworkers, gresources, workerID)
        wresources.set_rset_team([wrk])
        _worker_asserts(wresources, exp_out, exp_slots, wrk, nworkers, 1, 2)
        del wresources

    os.remove('node_list')


def test_get_local_nodelist_distrib_mode_host_not_in_list():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0036,0137-0139,1234]"
    resource_info = {'nodelist_env_slurm': "LIBE_RESOURCES_TEST_NODE_LIST"}
    libE_specs = {'resource_info': resource_info,
                  'dedicated_mode': False}

    gresources = GlobalResources(libE_specs)
    nworkers = 4
    exp_out = ['knl-0022', 'knl-0036']

    # Test running distributed mode without current host in list.
    workerID = 2
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    assert wresources.local_nodelist == exp_out, "local_nodelist returned does not match expected"
    del wresources


def test_get_local_nodelist_distrib_mode():
    mynode = socket.gethostname()
    # nodelist_in = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    nodelist_in = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036', 'knl-0137', 'knl-0138', 'knl-0139']
    with open('node_list', 'w') as f:
        for i, node in enumerate(nodelist_in):
            f.write(node + '\n')
            if i == 3:
                f.write(mynode + '\n')

    libE_specs = {'dedicated_mode': False}
    gresources = GlobalResources(libE_specs)
    gresources.add_comm_info(libE_nodes=[mynode])

    # Spoof current process as each worker and check nodelist.
    nworkers = 8
    workerID = 5
    exp_node = mynode  # sname(mynode)
    exp_out = [exp_node]
    exp_slots = {exp_node: [0]}
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    assert wresources.local_nodelist == exp_out, "local_nodelist returned does not match expected"

    # Do the rest
    exp_split = [['knl-0020'], ['knl-0021'], ['knl-0022'], ['knl-0036'],
                 [exp_node], ['knl-0137'], ['knl-0138'], ['knl-0139']]
    _worker_asserts(wresources, exp_split, exp_slots, workerID-1, nworkers, 1)
    del wresources

    nworkers = 1
    workerID = 1
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    exp_out = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036', exp_node, 'knl-0137', 'knl-0138', 'knl-0139']
    assert wresources.local_nodelist == exp_out, "local_nodelist returned does not match expected"
    del wresources

    nworkers = 4
    workerID = 3
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    exp_out = [exp_node, 'knl-0137']
    assert wresources.local_nodelist == exp_out, "local_nodelist returned does not match expected"
    del wresources

    # Sub-node workers
    nworkers = 16
    workerID = 9
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    exp_out = [exp_node]
    assert wresources.local_nodelist == exp_out, "local_nodelist returned does not match expected"
    del wresources

    workerID = 10
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    exp_out = [exp_node]
    assert wresources.local_nodelist == exp_out, "local_nodelist returned does not match expected"
    del wresources

    os.remove('node_list')


def test_get_local_nodelist_distrib_mode_uneven_split():
    mynode = socket.gethostname()
    exp_node = mynode  # sname(mynode)
    nodelist_in = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    with open('node_list', 'w') as f:
        for i, node in enumerate(nodelist_in):
            f.write(node + '\n')
            if i == 4:
                f.write(mynode + '\n')

    libE_specs = {'dedicated_mode': False}
    gresources = GlobalResources(libE_specs)
    gresources.add_comm_info(libE_nodes=[mynode])
    nworkers = 2

    # May not be at head of list - should perhaps be warning or enforced
    exp_out_w1 = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0036', 'knl-0137']
    exp_out_w2 = [exp_node, 'knl-0138', 'knl-0139', 'knl-1234']
    exp_split = [exp_out_w1, exp_out_w2]

    workerID = 2
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    exp_slots = {exp_node: [0], 'knl-0138': [0], 'knl-0139': [0], 'knl-1234': [0]}
    _worker_asserts(wresources, exp_split, exp_slots, workerID-1, nworkers, 4)
    del wresources

    workerID = 1
    wresources = WorkerResources(nworkers, gresources, workerID)
    wresources.set_rset_team([workerID-1])
    exp_slots = {'knl-0020': [0], 'knl-0021': [0], 'knl-0022': [0], 'knl-0036': [0], 'knl-0137': [0]}
    _worker_asserts(wresources, exp_split, exp_slots, workerID-1, nworkers, 5)
    del wresources

    os.remove('node_list')


def test_map_workerid_to_index():
    num_workers = 4
    num_rsets = 4

    zero_resource_list = []
    index_list = ResourceManager.get_index_list(num_workers, num_rsets, zero_resource_list)
    for workerID in range(1, num_workers+1):
        index = index_list[workerID-1]
        assert index == workerID - 1, "index incorrect. Received: " + str(index)

    zero_resource_list = [1]
    index_list = ResourceManager.get_index_list(num_workers, num_rsets, zero_resource_list)
    for workerID in range(2, num_workers+1):
        index = index_list[workerID-1]
        assert index == workerID - 2, "index incorrect. Received: " + str(index)

    zero_resource_list = [1, 2]
    index_list = ResourceManager.get_index_list(num_workers, num_rsets, zero_resource_list)
    for workerID in range(3, num_workers+1):
        index = index_list[workerID-1]
        assert index == workerID - 3, "index incorrect. Received: " + str(index)

    zero_resource_list = [1, 3]
    index_list = ResourceManager.get_index_list(num_workers, num_rsets, zero_resource_list)

    workerID = 2
    index = index_list[workerID-1]
    assert index == 0, "index incorrect. Received: " + str(index)

    workerID = 4
    index = index_list[workerID-1]
    assert index == 1, "index incorrect. Received: " + str(index)


def _check_mfile(machinefile, exp_list):
    with open('machinefile', 'r') as f:
        i = 0
        for line in f:
            index = i//4
            assert line == exp_list[index]
            i += 1


def test_machinefile_from_resources():

    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0036,0137-0139,1234]"
    resource_info = {'nodelist_env_slurm': "LIBE_RESOURCES_TEST_NODE_LIST"}
    libE_specs = {'resource_info': resource_info,
                  'num_resource_sets': 8}

    exp_list = ['knl-0020\n', 'knl-0021\n', 'knl-0022\n', 'knl-0036\n']

    resources = Resources(libE_specs)
    resources.set_worker_resources(4, 1)
    resources.worker_resources.set_rset_team([0, 1, 2, 3])

    built_mfile = create_machinefile(resources, num_nodes=4, procs_per_node=4)
    assert built_mfile, \
        "machinefile doesn't exist or is empty"

    _check_mfile('machinefile', exp_list)

    # Test replacing older machinefile
    create_machinefile(resources, machinefile='machinefile', num_nodes=4, procs_per_node=4)
    _check_mfile('machinefile', exp_list)
    os.remove('machinefile')


if __name__ == "__main__":
    setup_standalone_run()

    test_get_global_nodelist_frm_slurm()
    test_get_global_nodelist_frm_slurm_suffix()
    test_get_global_nodelist_frm_slurm_single()
    test_get_global_nodelist_frm_slurm_straight_list()
    test_get_global_nodelist_frm_slurm_multigroup()
    test_get_global_nodelist_frm_slurm_complex()

    test_get_global_nodelist_frm_cobalt()
    test_get_global_nodelist_frm_lsf()
    test_get_global_nodelist_frm_lsf_shortform()
    test_get_global_nodelist_standalone()

    test_get_global_nodelist_frm_wrklst_file()
    test_remove_libE_nodes()

    test_get_local_resources_dedicated_mode()
    test_get_local_resources_dedicated_mode_remove_libE_proc()
    test_get_local_nodelist_distrib_mode_host_not_in_list()
    test_get_local_nodelist_distrib_mode()
    test_get_local_nodelist_distrib_mode_uneven_split()

    test_map_workerid_to_index()
    test_machinefile_from_resources()

    teardown_standalone_run()



# """
import numpy as np
import math


# Concave multiobjective function with the shape of the unit sphere in the
# first orthant. Credit to Deb et al. [2002].
def dtlz2(x, f):
    d = len(x)
    p = len(f)
    # Compute the kernel function g(x)
    gx = np.dot(x[p-1:d]-0.5, x[p-1:d]-0.5)
    # Compute the objectives
    f[0] = (1.0 + gx)
    for y in x[:p-1]:
        f[0] *= math.cos(math.pi * y / 2.0)
    for i in range(1, p):
        f[i] = (1.0 + gx) * math.sin(math.pi * x[p-1-i] / 2.0)
        for y in x[:p-1-i]:
            f[i] *= math.cos(math.pi * y / 2.0)
    return


# Convex multiobjective function with parabolic Pareto front
def convex_mop(x, f):
    d = len(x)
    p = len(f)
    for i in range(p):
        e = np.zeros(d)
        e[i] = 0.5
        f[i] = np.dot(x-e, x-e)
    return



"""
Test the use for APOSMM with an external local optimization method.
Points are passed to/from the localopt method using files with run-specific
hashes. These hashes are currently generated using uuid, which may not be
thread safe on some systems (e.g., Travis-CI). This was resolved by not using
'local' communication; we therefore recommend using 'mpi' communication when
using persistent_aposmm with an external localopt # method.


Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_external_localopt.py

When running with the above command, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.

"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi tcp
# TESTSUITE_NPROCS: 4
# TESTSUITE_OS_SKIP: OSX
# TESTSUITE_EXTRA: true

import sys
import numpy as np
import shutil  # For copying the external_localopt script

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = 'nlopt'
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f

from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.tests.regression_tests.support import six_hump_camel_minima as minima
from time import time

np.set_printoptions(precision=16)

nworkers, is_manager, libE_specs, _ = parse_args()

if is_manager:
    start_time = time()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_out = [
    ('x', float, n),
    ('x_on_cube', float, n),
    ('sim_id', int),
    ('local_min', bool),
    ('local_pt', bool),
]

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'sample_points': np.round(minima, 1),
        'localopt_method': 'external_localopt',
        'max_active_runs': 6,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}
shutil.copy('./scripts_used_by_reg_tests/call_matlab_octave_script.m', './')
shutil.copy('./scripts_used_by_reg_tests/wrapper_obj_fun.m', './')

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 500}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    print('[Manager]:', H[np.where(H['local_min'])]['x'])
    print('[Manager]: Time taken =', time() - start_time, flush=True)

    # Note: This regression test considers only the global minima because it's
    # not possible to pass an initial simplex to fminsearch in Octave/Matlab.
    # As such, localopt runs in APOSMM that are started near 4 of the 6 local
    # minima "jump out", even when 'sim_max' is increased. (Matlab's fminsearch
    # has a smaller initial simplex and appears to be less susceptible to
    # this.)
    minima = minima[:2]
    tol = 1e-3
    for m in minima:
        # The minima are known on this test problem.
        # We use their values to test APOSMM has identified all minima
        print(np.min(np.sum((H[H['local_min']]['x'] - m) ** 2, 1)), flush=True)
        assert np.min(np.sum((H[H['local_min']]['x'] - m) ** 2, 1)) < tol

    save_libE_output(H, persis_info, __file__, nworkers)



"""
Manages libensemble resources related to MPI tasks launched from nodes.
"""

import os
import logging
import subprocess


class MPIResourcesException(Exception):
    "Resources module exception."


def rassert(test, *args):
    if not test:
        raise MPIResourcesException(*args)


logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


def get_MPI_runner():
    """ Return whether ``mpirun`` is openmpi or mpich """
    var = get_MPI_variant()
    if var in ['mpich', 'openmpi']:
        return 'mpirun'
    else:
        return var


def get_MPI_variant():
    """Returns MPI base implementation

    Returns
    -------
    mpi_variant: string:
        MPI variant 'aprun' or 'jsrun' or 'mpich' or 'openmpi' or 'srun'

    """

    try:
        subprocess.check_call(['aprun', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return 'aprun'
    except OSError:
        pass

    try:
        subprocess.check_call(['jsrun', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return 'jsrun'
    except OSError:
        pass

    try:
        # Explore mpi4py.MPI.get_vendor() and mpi4py.MPI.Get_library_version() for mpi4py
        try_mpich = subprocess.Popen(['mpirun', '-npernode'], stdout=subprocess.PIPE,
                                     stderr=subprocess.STDOUT)
        stdout, _ = try_mpich.communicate()
        if 'unrecognized argument npernode' in stdout.decode():
            return 'mpich'
        return 'openmpi'
    except Exception:
        pass

    try:
        subprocess.check_call(['srun', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return 'srun'
    except OSError:
        pass


def task_partition(num_procs, num_nodes, procs_per_node, machinefile=None):
    """Takes provided nprocs/nodes/ranks and outputs working
    configuration of procs/nodes/ranks or error
    """

    # Convert to int if string is provided
    num_procs = int(num_procs) if num_procs else None
    num_nodes = int(num_nodes) if num_nodes else None
    procs_per_node = int(procs_per_node) if procs_per_node else None

    # If machinefile is provided - ignore everything else
    if machinefile:
        if num_procs or num_nodes or procs_per_node:
            logger.warning("Machinefile provided - overriding "
                           "procs/nodes/procs_per_node")
        return None, None, None

    if not num_procs:
        rassert(num_nodes and procs_per_node,
                "Need num_procs, num_nodes/procs_per_node, or machinefile")
        num_procs = num_nodes * procs_per_node

    elif not num_nodes:
        procs_per_node = procs_per_node or num_procs
        num_nodes = num_procs//procs_per_node

    elif not procs_per_node:
        procs_per_node = num_procs//num_nodes

    rassert(num_procs == num_nodes*procs_per_node,
            "num_procs does not equal num_nodes*procs_per_node")
    return num_procs, num_nodes, procs_per_node


def _max_rsets_per_node(worker_resources):
    """ Return the maximum rsets per node for any node on this worker"""
    rset_team = worker_resources.rset_team
    local_rsets_list = worker_resources.local_rsets_list
    rsets_on_node = [local_rsets_list[rset] for rset in rset_team]
    return max(rsets_on_node)


def get_resources(resources, num_procs=None, num_nodes=None,
                  procs_per_node=None, hyperthreads=False):
    """Reconciles user-supplied options with available worker
    resources to produce run configuration.

    Detects resources available to worker, checks whether an existing
    user-supplied config is valid, and fills in any missing config
    information (i.e., num_procs/num_nodes/procs_per_node)

    User-supplied config options are honored, and an exception is
    raised if these are infeasible.
    """
    wresources = resources.worker_resources
    gresources = resources.glob_resources
    node_list = wresources.local_nodelist
    rassert(node_list, "Node list is empty - aborting")
    local_node_count = wresources.local_node_count

    cores_avail_per_node = \
        (gresources.logical_cores_avail_per_node if hyperthreads else
         gresources.physical_cores_avail_per_node)

    rsets_per_node = _max_rsets_per_node(wresources)

    # Advantage of cores per rset first is they will always get same no. per rset (double rsets, double cores)
    # Advantage of multiply first is less wasted cores.
    cores_avail_per_node_per_worker = cores_avail_per_node//rsets_per_node * wresources.slot_count
    # cores_avail_per_node_per_worker = int(cores_avail_per_node/rsets_per_node * wresources.slot_count)

    rassert(wresources.even_slots,
            "Uneven distribution of node resources not yet supported. Nodes and slots are: {}"
            .format(wresources.slots))

    if not num_procs and not procs_per_node:
        rassert(cores_avail_per_node_per_worker > 0,
                "There is less than one core per resource set. "
                "Provide num_procs or num_nodes/procs_per_node to oversubsribe")
        procs_per_node = cores_avail_per_node_per_worker
        if not num_nodes:
            # If no decomposition supplied - use all available cores/nodes
            num_nodes = local_node_count
            logger.debug("No decomposition supplied - "
                         "using all available resource. "
                         "Nodes: {}  procs_per_node {}".
                         format(num_nodes, procs_per_node))
    elif not num_nodes and not procs_per_node:
        if num_procs <= cores_avail_per_node_per_worker:
            num_nodes = 1
        else:
            num_nodes = local_node_count
    elif not num_procs and not num_nodes:
        num_nodes = local_node_count

    # Checks config is consistent and sufficient to express
    num_procs, num_nodes, procs_per_node = \
        task_partition(num_procs, num_nodes, procs_per_node)

    rassert(num_nodes <= local_node_count,
            "Not enough nodes to honor arguments. "
            "Requested {}. Only {} available".
            format(num_nodes, local_node_count))

    if gresources.enforce_worker_core_bounds:
        rassert(procs_per_node <= cores_avail_per_node,
                "Not enough processors on a node to honor arguments. "
                "Requested {}. Only {} available".
                format(procs_per_node, cores_avail_per_node))

        rassert(procs_per_node <= cores_avail_per_node_per_worker,
                "Not enough processors per worker to honor arguments. "
                "Requested {}. Only {} available".
                format(procs_per_node, cores_avail_per_node_per_worker))

        rassert(num_procs <= (cores_avail_per_node * local_node_count),
                "Not enough procs to honor arguments. "
                "Requested {}. Only {} available".
                format(num_procs, cores_avail_per_node*local_node_count))

    if num_nodes < local_node_count:
        logger.warning("User constraints mean fewer nodes being used "
                       "than available. {} nodes used. {} nodes available".
                       format(num_nodes, local_node_count))

    return num_procs, num_nodes, procs_per_node


def create_machinefile(resources, machinefile=None, num_procs=None,
                       num_nodes=None, procs_per_node=None,
                       hyperthreads=False):
    """Creates a machinefile based on user-supplied config options,
    completed by detected machine resources
    """

    machinefile = machinefile or 'machinefile'
    if os.path.isfile(machinefile):
        try:
            os.remove(machinefile)
        except Exception as e:
            logger.warning("Could not remove existing machinefile: {}".format(e))

    node_list = resources.worker_resources.local_nodelist
    logger.debug("Creating machinefile with {} nodes and {} ranks per node".
                 format(num_nodes, procs_per_node))

    with open(machinefile, 'w') as f:
        for node in node_list[:num_nodes]:
            f.write((node + '\n') * procs_per_node)

    built_mfile = (os.path.isfile(machinefile)
                   and os.path.getsize(machinefile) > 0)
    return built_mfile, num_procs, num_nodes, procs_per_node


def get_hostlist(resources, num_nodes=None):
    """Creates a hostlist based on user-supplied config options.

    completed by detected machine resources
    """
    node_list = resources.worker_resources.local_nodelist
    hostlist_str = ",".join([str(x) for x in node_list[:num_nodes]])
    return hostlist_str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This is for setting up libEnsemble, license and details can be
# found at https://github.com/Libensemble/libensemble/

from setuptools import setup
from setuptools.command.test import test as TestCommand

exec(open('libensemble/version.py').read())


class Run_TestSuite(TestCommand):
    def run_tests(self):
        import os
        import sys
        py_version = sys.version_info[0]
        print('Python version from setup.py is', py_version)
        run_string = "libensemble/tests/run-tests.sh -p " + str(py_version)
        os.system(run_string)


class ToxTest(TestCommand):
    user_options = []

    def initialize_options(self):
        TestCommand.initialize_options(self)

    def run_tests(self):
        import tox
        tox.cmdline()


setup(
    name='libensemble',
    version=__version__,
    description='Library to coordinate the concurrent evaluation of dynamic ensembles of calculations',
    url='https://github.com/Libensemble/libensemble',
    author='Jeffrey Larson, Stephen Hudson, Stefan M. Wild, David Bindel and John-Luke Navarro',
    author_email='libensemble@lists.mcs.anl.gov',
    license='BSD 3-clause',

    packages=['libensemble',
              'libensemble.gen_funcs',
              'libensemble.sim_funcs',
              'libensemble.sim_funcs.branin',
              'libensemble.alloc_funcs',
              'libensemble.tests',
              'libensemble.comms',
              'libensemble.utils',
              'libensemble.tools',
              'libensemble.executors',
              'libensemble.resources',
              'libensemble.tests.unit_tests',
              'libensemble.tests.regression_tests'],

    package_data={'libensemble.sim_funcs.branin': ['known_minima_and_func_values']},

    install_requires=['numpy', 'psutil'],

    # If run tests through setup.py - downloads these but does not install
    tests_require=['pytest>=3.1',
                   'pytest-cov>=2.5',
                   'pytest-pep8>=1.0',
                   'pytest-timeout',
                   'mock'
                   ],

    extras_require={
        'extras': ['pyyaml', 'scipy', 'nlopt', 'mpi4py', 'petsc', 'petsc4py', 'DFO-LS', 'deap', 'mpmath'],
        'docs': ['sphinx', 'sphinxcontrib.bibtex', 'sphinx_rtd_theme']},

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],

    cmdclass={'test': Run_TestSuite, 'tox': ToxTest}
)



#!/usr/bin/env python
"""
A small example of using the call_branin function from branin_obj.py in order
to (write to)/(read from) files when doing a sim_f evaluation
"""

import numpy as np


def branin(x1, x2):
    a = 1.0
    b = 5.1/(4.0*pow(np.pi, 2.0))
    c = 5.0/np.pi
    r = 6.0
    s = 10.0
    t = 1.0/(8.0*np.pi)

    f = a*(x2 - b*x1**2 + c*x1 - r)**2 + s*(1-t)*np.cos(x1) + s

    return f


if __name__ == "__main__":

    x = np.loadtxt('./x.in')
    assert len(x) == 2, "branin.py is two dimensional"

    f = branin(x[0], x[1])
    np.savetxt('f.out', [f])



"""
This module contains various versions that evaluate the six-hump camel function.

Six-hump camel function is documented here:
  https://www.sfu.ca/~ssurjano/camel6.html

"""
__all__ = ['six_hump_camel', 'six_hump_camel_simple',
           'six_hump_camel_with_variable_resources',
           'six_hump_camel_CUDA_variable_resources',
           'persistent_six_hump_camel']

# import subprocess
import os
import sys
import numpy as np
import time
from libensemble.executors.executor import Executor
from libensemble.message_numbers import UNSET_TAG, WORKER_DONE, TASK_FAILED
from libensemble.resources.resources import Resources
from libensemble.tools.persistent_support import PersistentSupport
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, EVAL_SIM_TAG, FINISHED_PERSISTENT_SIM_TAG


def six_hump_camel(H, persis_info, sim_specs, _):
    """
    Evaluates the six hump camel function for a collection of points given in ``H['x']``.
    Additionally evaluates the gradient if ``'grad'`` is a field in
    ``sim_specs['out']`` and pauses for ``sim_specs['user']['pause_time']]`` if
    defined.

    .. seealso::
        `test_old_aposmm_with_gradients.py  <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_old_aposmm_with_gradients.py>`_ # noqa
    """

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        H_o['f'][i] = six_hump_camel_func(x)

        if 'grad' in H_o.dtype.names:
            H_o['grad'][i] = six_hump_camel_grad(x)

        if 'user' in sim_specs and 'pause_time' in sim_specs['user']:
            time.sleep(sim_specs['user']['pause_time'])

    return H_o, persis_info


def six_hump_camel_simple(x, persis_info, sim_specs, _):
    """
    Evaluates the six hump camel function for a single point ``x``.

    .. seealso::
        `test_fast_alloc.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_fast_alloc.py>`_ # noqa
    """

    H_o = np.zeros(1, dtype=sim_specs['out'])

    H_o['f'] = six_hump_camel_func(x[0][0])

    if 'pause_time' in sim_specs['user']:
        time.sleep(sim_specs['user']['pause_time'])

    return H_o, persis_info


def six_hump_camel_with_variable_resources(H, persis_info, sim_specs, libE_info):
    """
    Evaluates the six hump camel for a collection of points given in ``H['x']``
    via the executor, supporting variable sized simulations/resources, as
    determined by the generator.

    .. seealso::
        `test_uniform_sampling_with_variable_resources.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_with_variable_resources.py>`_ # noqa
    """

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])
    app = sim_specs['user'].get('app', 'helloworld')
    dry_run = sim_specs['user'].get('dry_run', False)  # dry_run only prints run lines in ensemble.log
    set_cores_by_rsets = True  # If True use rset count to set num procs, else use all available to this worker.
    core_multiplier = 1  # Only used with set_cores_by_rsets as a multiplier.

    exctr = Executor.executor  # Get Executor
    task_states = []
    for i, x in enumerate(H['x']):
        # If set_cores_by_rsets is set - set nprocs here (you can oversubscribe on node)
        # else let automatic resources set nodes/procs to use all available set.
        nprocs = None  # Will be as if argument is not present
        if set_cores_by_rsets:
            resources = Resources.resources.worker_resources
            nprocs = resources.num_rsets * core_multiplier
            # print('nprocs is',nprocs,flush=True)

        inpt = None  # Will be as if argument is not present
        if app == 'six_hump_camel':
            inpt = ' '.join(map(str, H['x'][i]))

        task = exctr.submit(app_name=app, app_args=inpt,
                            num_procs=nprocs,
                            stdout='out.txt', stderr='err.txt',
                            dry_run=dry_run)
        task.wait()
        # while(not task.finished):
        #     time.sleep(0.1)
        #     task.poll()

        task_states.append(task.state)

        if app == 'six_hump_camel':
            # H_o['f'][i] = float(task.read_stdout())  # Reads whole file
            with open('out.txt') as f:
                H_o['f'][i] = float(f.readline().strip())  # Read just first line
        else:
            # To return something in test
            H_o['f'][i] = six_hump_camel_func(x)

        # v = np.random.uniform(0, 10)
        # print('About to sleep for :' + str(v))
        # time.sleep(v)

    calc_status = UNSET_TAG  # Returns to worker
    if all(t == 'FINISHED' for t in task_states):
        calc_status = WORKER_DONE
    elif any(t == 'FAILED' for t in task_states):
        calc_status = TASK_FAILED

    return H_o, persis_info, calc_status


def six_hump_camel_CUDA_variable_resources(H, persis_info, sim_specs, libE_info):
    """Launches an app setting GPU resources

    The standard test apps do not run on GPU, but demonstrates accessing resource
    information to set ``CUDA_VISIBLE_DEVICES``, and typical run configuration.
    """
    x = H['x'][0]
    H_o = np.zeros(1, dtype=sim_specs['out'])

    # Interrogate resources available to this worker
    resources = Resources.resources.worker_resources
    if resources.even_slots:  # Need same slots on each node
        resources.set_env_to_slots("CUDA_VISIBLE_DEVICES")  # Use convenience function.
        num_nodes = resources.local_node_count
        cores_per_node = resources.slot_count  # One CPU per GPU
        print('CUDA_VISIBLE_DEVICES={}  \tnodes {} ppn {}'
              .format(os.environ["CUDA_VISIBLE_DEVICES"], num_nodes, cores_per_node), flush=True)
    else:
        # Unless use a matching sub-set, but usually you probably don't want this
        print('Error: Cannot set CUDA_VISIBLE_DEVICES when uneven slots on nodes {}'.format(resources.slots))

    # Create application input file
    inpt = ' '.join(map(str, x))
    exctr = Executor.executor  # Get Executor

    # Launch application via system MPI runner, using assigned resources.
    task = exctr.submit(app_name='six_hump_camel',
                        app_args=inpt,
                        num_nodes=num_nodes,
                        procs_per_node=cores_per_node,
                        stdout='out.txt',
                        stderr='err.txt')

    task.wait()  # Wait for run to complete

    # Access app output
    with open('out.txt') as f:
        H_o['f'] = float(f.readline().strip())  # Read just first line

    calc_status = WORKER_DONE if task.state == 'FINISHED' else 'FAILED'
    return H_o, persis_info, calc_status


def persistent_six_hump_camel(H, persis_info, sim_specs, libE_info):
    """
    Similar to ``six_hump_camel``, but runs in persistent mode.
    """

    ps = PersistentSupport(libE_info, EVAL_SIM_TAG)

    # Either start with a work item to process - or just start and wait for data
    if H.size > 0:
        tag = None
        Work = None
        calc_in = H
    else:
        tag, Work, calc_in = ps.recv()

    while tag not in [STOP_TAG, PERSIS_STOP]:

        # calc_in: This should either be a function (unpack_work ?) or included/unpacked in ps.recv/ps.send_recv.
        if Work is not None:
            persis_info = Work.get('persis_info', persis_info)
            libE_info = Work.get('libE_info', libE_info)

        # Call standard six_hump_camel sim
        H_o, persis_info = six_hump_camel(calc_in, persis_info, sim_specs, libE_info)

        tag, Work, calc_in = ps.send_recv(H_o)

    final_return = None

    # Overwrite final point - for testing only
    if sim_specs['user'].get('replace_final_fields', 0):
        calc_in = np.ones(1, dtype=[('x', float, (2,))])
        H_o, persis_info = six_hump_camel(calc_in, persis_info, sim_specs, libE_info)
        final_return = H_o

    return final_return, persis_info, FINISHED_PERSISTENT_SIM_TAG


def six_hump_camel_func(x):
    """
    Definition of the six-hump camel
    """
    x1 = x[0]
    x2 = x[1]
    term1 = (4-2.1*x1**2+(x1**4)/3) * x1**2
    term2 = x1*x2
    term3 = (-4+4*x2**2) * x2**2

    return term1 + term2 + term3


def six_hump_camel_grad(x):
    """
    Definition of the six-hump camel gradient
    """

    x1 = x[0]
    x2 = x[1]
    grad = np.zeros(2)

    grad[0] = 2.0*(x1**5 - 4.2*x1**3 + 4.0*x1 + 0.5*x2)
    grad[1] = x1 + 16*x2**3 - 8*x2

    return grad


if __name__ == "__main__":
    x = (float(sys.argv[1]), float(sys.argv[2]))
    result = six_hump_camel_func(x)
    print(result)



"""
Tests libEnsemble's capability to kill/cancel  simulations that are in progress.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_surmise_killsims.py
   python3 test_persistent_surmise_killsims.py --nworkers 3 --comms local
   python3 test_persistent_surmise_killsims.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.

This test is a smaller variant of test_persistent_surmise_calib.py, but which
subprocesses a compiled version of the borehole simulation. A delay is
added to simulations after the initial batch, so that the killing of running
simulations can be tested. This will only affect simulations that have already
been issued to a worker when the cancel request is registesred by the manager.

See more information, see tutorial:
"Borehole Calibration with Selective Simulation Cancellation"
in the libEnsemble documentation.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4
# TESTSUITE_EXTRA: true

# Requires:
#   Install Surmise package

import os
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.gen_funcs.persistent_surmise_calib import surmise_calib as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.sim_funcs.borehole_kills import borehole as sim_f
from libensemble.tests.regression_tests.common import build_borehole  # current location
from libensemble.executors.executor import Executor
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

# from libensemble import logger
# logger.set_level('DEBUG')  # To get debug logging in ensemble.log

if __name__ == '__main__':

    nworkers, is_manager, libE_specs, _ = parse_args()

    n_init_thetas = 15  # Initial batch of thetas
    n_x = 5  # No. of x values
    nparams = 4  # No. of theta params
    ndims = 3  # No. of x coordinates.
    max_add_thetas = 20  # Max no. of thetas added for evaluation
    step_add_theta = 10  # No. of thetas to generate per step, before emulator is rebuilt
    n_explore_theta = 200  # No. of thetas to explore while selecting the next theta
    obsvar = 10 ** (-1)  # Constant for generating noise in obs

    # Batch mode until after init_sample_size (add one theta to batch for observations)
    init_sample_size = (n_init_thetas + 1) * n_x

    # Stop after max_emul_runs runs of the emulator
    max_evals = init_sample_size + max_add_thetas * n_x

    sim_app = os.path.join(os.getcwd(), "borehole.x")
    if not os.path.isfile(sim_app):
        build_borehole()

    exctr = Executor()  # Run serial sub-process in place
    exctr.register_app(full_path=sim_app, app_name='borehole')

    # Subprocess variant creates input and output files for each sim
    libE_specs['sim_dirs_make'] = True  # To keep all - make sim dirs
    # libE_specs['use_worker_dirs'] = True  # To overwrite - make worker dirs only

    # Rename ensemble dir for non-inteference with other regression tests
    libE_specs['ensemble_dir_path'] = 'ensemble_calib_kills'

    sim_specs = {
        'sim_f': sim_f,
        'in': ['x', 'thetas'],
        'out': [('f', float)],
        'user': {
            'num_obs': n_x,
            'init_sample_size': init_sample_size,
        },
    }

    gen_out = [
        ('x', float, ndims),
        ('thetas', float, nparams),
        ('priority', int),
        ('obs', float, n_x),
        ('obsvar', float, n_x),
    ]

    gen_specs = {
        'gen_f': gen_f,
        'persis_in': [o[0] for o in gen_out] + ['f', 'returned', 'sim_id'],
        'out': gen_out,
        'user': {
            'n_init_thetas': n_init_thetas,  # Num thetas in initial batch
            'num_x_vals': n_x,  # Num x points to create
            'step_add_theta': step_add_theta,  # No. of thetas to generate per step
            'n_explore_theta': n_explore_theta,  # No. of thetas to explore each step
            'obsvar': obsvar,  # Variance for generating noise in obs
            'init_sample_size': init_sample_size,  # Initial batch size inc. observations
            'priorloc': 1,  # Prior location in the unit cube.
            'priorscale': 0.2,  # Standard deviation of prior
        },
    }

    alloc_specs = {
        'alloc_f': alloc_f,
        'user': {
            'init_sample_size': init_sample_size,
            'async_return': True,  # True = Return results to gen as they come in (after sample)
            'active_recv_gen': True,  # Persistent gen can handle irregular communications
        },
    }

    persis_info = add_unique_random_streams({}, nworkers + 1)
    exit_criteria = {'sim_max': max_evals}

    # Perform the run
    H, persis_info, flag = libE(
        sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs=alloc_specs, libE_specs=libE_specs
    )

    if is_manager:
        print('Cancelled sims', H['sim_id'][H['cancel_requested']])
        print('Killed sims', H['sim_id'][H['kill_sent']])
        sims_done = np.count_nonzero(H['returned'])
        save_libE_output(H, persis_info, __file__, nworkers)
        assert sims_done == max_evals, 'Num of completed simulations should be {}. Is {}'.format(max_evals, sims_done)



"""
Common plumbing for regression tests
"""

import os
import json
import os.path


def create_node_file(num_nodes, name='node_list'):
    """Create a nodelist file"""
    if os.path.exists(name):
        os.remove(name)
    with open(name, 'w') as f:
        for i in range(1, num_nodes + 1):
            f.write('node-' + str(i) + '\n')
        f.flush()
        os.fsync(f)


def mpi_comm_excl(exc=[0], comm=None):
    "Exclude ranks from a communicator for MPI comms."
    from mpi4py import MPI

    parent_comm = comm or MPI.COMM_WORLD
    parent_group = parent_comm.Get_group()
    new_group = parent_group.Excl(exc)
    mpi_comm = parent_comm.Create(new_group)
    return mpi_comm, MPI.COMM_NULL


def mpi_comm_split(num_parts, comm=None):
    "Split COMM_WORLD into sub-communicators for MPI comms."
    from mpi4py import MPI

    parent_comm = comm or MPI.COMM_WORLD
    parent_size = parent_comm.Get_size()
    key = parent_comm.Get_rank()
    row_size = parent_size // num_parts
    sub_comm_number = key // row_size
    sub_comm = parent_comm.Split(sub_comm_number, key)
    return sub_comm, sub_comm_number


def build_simfunc():
    import subprocess

    # Build simfunc
    # buildstring='mpif90 -o my_simtask.x my_simtask.f90' # On cray need to use ftn
    buildstring = 'mpicc -o my_simtask.x ../unit_tests/simdir/my_simtask.c'
    # subprocess.run(buildstring.split(),check=True) #Python3.5+
    subprocess.check_call(buildstring.split())


def build_borehole():
    import subprocess

    buildstring = 'gcc -o borehole.x ../unit_tests/simdir/borehole.c -lm'
    subprocess.check_call(buildstring.split())


def modify_Balsam_worker():
    # Balsam is meant for HPC systems that commonly distribute jobs across many
    #   nodes. Due to the nature of testing Balsam on local or CI systems which
    #   usually only contain a single node, we need to change Balsam's default
    #   worker setup so multiple workers can be run on a single node.
    import balsam

    new_lines = [
        "        for idx in range(10):\n",
        "            w = Worker(1, host_type='DEFAULT', num_nodes=1)\n",
        "            self.workers.append(w)\n",
    ]

    workerfile = 'worker.py'
    balsam_path = os.path.dirname(balsam.__file__) + '/launcher'
    balsam_worker_path = os.path.join(balsam_path, workerfile)

    with open(balsam_worker_path, 'r') as f:
        lines = f.readlines()

    if lines[-3] != new_lines[0]:
        lines = lines[:-2]  # effectively inserting new_lines[0] above
        lines.extend(new_lines)

    with open(balsam_worker_path, 'w') as f:
        for line in lines:
            f.write(line)


def modify_Balsam_pyCoverage():
    # Tracking line coverage through our tests requires running the Python module
    #   'coverage' directly. Balsam explicitly configures Python runs with
    #   'python [script].py args' with no current capability for specifying
    #   modules. This hack specifies the coverage module and some options.
    import balsam

    rcfile = os.path.abspath('./libensemble/tests/regression_tests/.bal_coveragerc')

    old_line = "            path = ' '.join((exe, script_path, args))\n"
    new_line = (
        "            path = ' '.join((exe, '-m coverage run "
        + "--parallel-mode --rcfile={}', script_path, args))\n".format(rcfile)
    )

    commandfile = 'cli_commands.py'
    balsam_path = os.path.dirname(balsam.__file__) + '/scripts'
    balsam_commands_path = os.path.join(balsam_path, commandfile)

    with open(balsam_commands_path, 'r') as f:
        lines = f.readlines()

    for i in range(len(lines)):
        if lines[i] == old_line:
            lines[i] = new_line

    with open(balsam_commands_path, 'w') as f:
        for line in lines:
            f.write(line)


def modify_Balsam_settings():
    # Set $HOME/.balsam/settings.json to DEFAULT instead of Theta worker setup
    settingsfile = os.path.join(os.environ.get('HOME'), '.balsam/settings.json')
    with open(settingsfile, 'r') as f:
        lines = json.load(f)

    lines['MPI_RUN_TEMPLATE'] = "MPICHCommand"
    lines['WORKER_DETECTION_TYPE'] = "DEFAULT"

    with open(settingsfile, 'w') as f:
        json.dump(lines, f)


def modify_Balsam_JobEnv():
    # If Balsam detects that the system on which it is running contains the string
    #   'cc' in its hostname, then it thinks it's on Cooley! If hostnames are
    #   randomly generated and occasionally may contain that offending string, then this
    #   modifies Balsam's JobEnvironment class to not check for 'cc'.
    import balsam

    bad_line = "        'COOLEY' : 'cooley cc'.split()\n"
    new_line = "        'COOLEY' : 'cooley'.split()\n"

    jobenv_file = 'JobEnvironment.py'
    balsam_path = os.path.dirname(balsam.__file__) + '/service/schedulers'
    balsam_jobenv_path = os.path.join(balsam_path, jobenv_file)

    with open(balsam_jobenv_path, 'r') as f:
        lines = f.readlines()

    for i in range(len(lines)):
        if lines[i] == bad_line:
            lines[i] = new_line

    with open(balsam_jobenv_path, 'w') as f:
        for line in lines:
            f.write(line)



import os
import mpi4py

path = mpi4py.__path__[0]
print("\nmpi4py path found is:", path)

configfile = os.path.join(path, "mpi.cfg")
print("\nShowing config file: ", configfile, "\n")

with open(configfile, 'r') as confile_handle:
    print(confile_handle.read())

with open(configfile, 'r') as infile:
    for line in infile:
        if line.startswith("mpicc ="):
            mpi4py_mpicc = line[8:-1]
            cmd_line = str(mpi4py_mpicc) + ' -v'
            break

print(cmd_line, ":\n")
os.system(cmd_line)



# """
# Runs libEnsemble with APOSMM+POUNDERS on the chwirut least squares problem.
# All 214 residual calculations for a given point are performed as a single
# simulation evaluation.
#
# Execute via one of the following commands (e.g. 3 workers):
#    mpiexec -np 4 python3 test_chwirut_pounders.py
#
# The number of concurrent evaluations of the objective function will be 4-1=3.
# """

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.chwirut1 import chwirut_eval as sim_f

import libensemble.gen_funcs
libensemble.gen_funcs.rc.aposmm_optimizers = 'petsc'
from libensemble.gen_funcs.old_aposmm import aposmm_logic as gen_f

from libensemble.tests.regression_tests.support import persis_info_2 as persis_info, aposmm_gen_out as gen_out
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

# Declare the run parameters/functions
m = 214
n = 3
budget = 10

sim_specs = {'sim_f': sim_f,
             'in': ['x'],
             'out': [('f', float), ('fvec', float, m)],
             'user': {'combine_component_func': lambda x: np.sum(np.power(x, 2))}
             }

gen_out += [('x', float, n), ('x_on_cube', float, n)]

# lb tries to avoid x[1]=-x[2], which results in division by zero in chwirut.
gen_specs = {'gen_f': gen_f,
             'in': [o[0] for o in gen_out]+['f', 'fvec', 'returned'],
             'out': gen_out,
             'user': {'initial_sample_size': 5,
                      'lb': (-2-np.pi/10)*np.ones(n),
                      'ub': 2*np.ones(n),
                      'localopt_method': 'pounders',
                      'dist_to_bound_multiple': 0.5,
                      'components': m}
             }

gen_specs['user'].update({'grtol': 1e-4, 'gatol': 1e-4, 'frtol': 1e-15, 'fatol': 1e-15})

persis_info = add_unique_random_streams(persis_info, nworkers + 1)

exit_criteria = {'sim_max': budget}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info,
                            libE_specs=libE_specs)

if is_manager:
    assert flag == 0
    assert len(H) >= budget

    # Calculating the Jacobian at the best point (though this information was not used by pounders)
    from libensemble.sim_funcs.chwirut1 import EvaluateJacobian
    J = EvaluateJacobian(H['x'][np.argmin(H['f'])])
    assert np.linalg.norm(J) < 2000

    save_libE_output(H, persis_info, __file__, nworkers)



"""
Created on Tue Feb  9 10:27:23 2021

@author: mosesyhc
"""
import numpy as np


def borehole(H, persis_info, sim_specs, libE_info):
    """
    Wraps the borehole function
    """

    H_o = np.zeros(H['x'].shape[0], dtype=sim_specs['out'])

    # If observation do not use failure model
    sim_id = libE_info['H_rows'][0]
    if sim_id > sim_specs['user']['num_obs']:
        H_o['f'] = borehole_failmodel(H['x'], H['thetas'])
    else:
        H_o['f'] = borehole_model(H['x'], H['thetas'])
    return H_o, persis_info


def borehole_failmodel(x, theta):
    """Given x and theta, return matrix of [row x] times [row theta] of values."""
    f = borehole_model(x, theta)
    wheretoobig = np.where((f / borehole_true(x)) > 1.25)
    f[wheretoobig[0], wheretoobig[1]] = np.inf
    return f


def borehole_model(x, theta):
    """Given x and theta, return matrix of [row x] times [row theta] of values."""
    theta = tstd2theta(theta)
    x = xstd2x(x)
    p = x.shape[0]
    n = theta.shape[0]

    theta_stacked = np.repeat(theta, repeats=p, axis=0)
    x_stacked = np.tile(x.astype(float), (n, 1))

    f = borehole_vec(x_stacked, theta_stacked).reshape((n, p))
    return f.T


def borehole_true(x):
    """Given x, return matrix of [row x] times 1 of values."""
    # assume true theta is [0.5]^d
    theta0 = np.atleast_2d(np.array([0.5] * 4))
    f0 = borehole_model(x, theta0)

    return f0


def borehole_vec(x, theta):
    """Given x and theta, return vector of values."""
    (Hu, Ld_Kw, Treff, powparam) = np.split(theta, theta.shape[1], axis=1)
    (rw, Hl) = np.split(x[:, :-1], 2, axis=1)
    numer = 2 * np.pi * (Hu - Hl)
    denom1 = 2 * Ld_Kw / rw ** 2
    denom2 = Treff

    f = ((numer / ((denom1 + denom2))) * np.exp(powparam * rw)).reshape(-1)

    return f


def tstd2theta(tstd, hard=True):
    """Given standardized theta in [0, 1]^d, return non-standardized theta."""
    if tstd.ndim < 1.5:
        tstd = tstd[:, None].T
    (Treffs, Hus, LdKw, powparams) = np.split(tstd, tstd.shape[1], axis=1)

    Treff = (0.5-0.05) * Treffs + 0.05
    Hu = Hus * (1110 - 990) + 990
    if hard:
        Ld_Kw = LdKw * (1680/1500 - 1120/15000) + 1120/15000
    else:
        Ld_Kw = LdKw*(1680/9855 - 1120/12045) + 1120/12045

    powparam = powparams * (0.5 - (- 0.5)) + (-0.5)

    theta = np.hstack((Hu, Ld_Kw, Treff, powparam))
    return theta


def xstd2x(xstd):
    """Given standardized x in [0, 1]^2 x {0, 1}, return non-standardized x."""
    if xstd.ndim < 1.5:
        xstd = xstd[:, None].T
    (rws, Hls, labels) = np.split(xstd, xstd.shape[1], axis=1)

    rw = rws * (np.log(0.5) - np.log(0.05)) + np.log(0.05)
    rw = np.exp(rw)
    Hl = Hls * (820 - 700) + 700

    x = np.hstack((rw, Hl, labels))
    return x



#!/usr/bin/env python

"""
Unit test of location stack for libensemble.
"""

import os
import shutil
import tempfile

from libensemble.utils.loc_stack import LocationStack


def test_location_stack():
    "Test correctness of location stack (all in a temp dir)."

    tmp_dirname = tempfile.mkdtemp()
    assert os.path.isdir(tmp_dirname), \
        "Failed to create temporary directory {}.".format(tmp_dirname)

    try:
        # Record where we started
        start_dir = os.getcwd()

        # Set up directory for clone
        clone_dirname = os.path.join(tmp_dirname, "basedir")
        os.mkdir(clone_dirname)
        test_fname = os.path.join(clone_dirname, "test.txt")
        with open(test_fname, "w+") as f:
            f.write("This is a test file\n")

        s = LocationStack()

        # Register a valid location
        tname = s.register_loc(0, "testdir",
                               prefix=tmp_dirname,
                               copy_files=[test_fname])
        assert os.path.isdir(tname), \
            "New directory {} was not created.".format(tname)
        assert os.path.isfile(os.path.join(tname, "test.txt")), \
            "New directory {} failed to copy test.txt from {}.". \
            format(tname, clone_dirname)

        # Register an empty location
        d = s.register_loc(1, None)
        assert d is None, \
            "Dir stack not correctly register None at location 1."

        # Register a dummy location (del should not work)
        d = s.register_loc(2, os.path.join(tmp_dirname, "dummy"))
        assert ~os.path.isdir(d), \
            "Directory stack registration of dummy should not create dir."

        # Push unregistered location (we should not move)
        s.push_loc(3)
        assert s.stack == [None], \
            "Directory stack push_loc(missing) failed to put None on stack."
        assert os.path.samefile(os.getcwd(), start_dir), \
            "Directory stack push_loc failed to stay put with input None." \
            "Wanted {}, at {}".format(start_dir, os.getcwd())

        # Push registered location (we should move
        s.push_loc(0)
        assert s.stack == [None, start_dir], \
            "Directory stack is incorrect." \
            "Wanted [None, {}], got {}.".format(start_dir, s.stack)
        assert os.path.samefile(os.getcwd(), tname), \
            "Directory stack push_loc failed to end up at desired dir." \
            "Wanted {}, at {}".format(tname, os.getcwd())

        # Pop the registered location
        s.pop()
        assert s.stack == [None], \
            "Directory stack is incorrect after pop." \
            "Wanted [None], got {}.".format(s.stack)
        assert os.path.samefile(os.getcwd(), start_dir), \
            "Directory stack push_loc failed to stay put with input None." \
            "Wanted {}, at {}".format(start_dir, os.getcwd())

        # Context for moving again
        with s.loc(0):
            assert s.stack == [None, start_dir], \
                "Directory stack is incorrect." \
                "Wanted [None, {}], got {}.".format(start_dir, s.stack)
            assert os.path.samefile(os.getcwd(), tname), \
                "Directory stack push_loc failed to end up at desired dir." \
                "Wanted {}, at {}".format(tname, os.getcwd())

        # Check directory after context
        assert s.stack == [None], \
            "Directory stack is incorrect after ctx." \
            "Wanted [None], got {}.".format(s.stack)
        assert os.path.samefile(os.getcwd(), start_dir), \
            "Directory looks wrong after ctx." \
            "Wanted {}, at {}".format(start_dir, os.getcwd())

        with s.dir(None):
            assert s.stack == [None, None], \
                "Directory stack is incorrect in ctx."
        assert s.stack == [None], \
            "Directory stack is incorrect after ctx."

        # Pop the unregistered location
        s.pop()
        assert not s.stack, \
            "Directory stack should be empty, actually {}.".format(s.stack)
        assert os.path.samefile(os.getcwd(), start_dir), \
            "Directory stack push_loc failed to stay put with input None." \
            "Wanted {}, at {}".format(start_dir, os.getcwd())

        # Clean up
        s.clean_locs()
        assert not os.path.isdir(tname), \
            "Directory {} should have been removed on cleanup.".format(tname)

    finally:
        shutil.rmtree(tmp_dirname)



"""
Tests libEnsemble's generator function requesting/receiving sim_f evaluations
asynchronously

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_uniform_sampling_async.py
   python3 test_persistent_uniform_sampling_async.py --nworkers 3 --comms local
   python3 test_persistent_uniform_sampling_async.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 4

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.branin.branin_obj import call_branin as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
    'user': {'uniform_random_pause_ub': 0.5},
}

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'x', 'sim_id'],
    'out': [('x', float, (n,))],
    'user': {
        'initial_batch_size': nworkers,  # Ensure > 1 alloc to send all sims
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {
    'alloc_f': alloc_f,
    'out': [],
    'user': {'async_return': True},
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'gen_max': 100, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    [_, counts] = np.unique(H['gen_time'], return_counts=True)
    print('Num. points in each gen iteration:', counts)
    assert counts[0] == nworkers, "The first gen_time should be common among initial_batch_size number of points"
    assert len(np.unique(counts)) > 1, "There is no variablitiy in the gen_times but there should be for the async case"

    save_libE_output(H, persis_info, __file__, nworkers)



#!/usr/bin/env python

"""
This file is part of the suite of scripts to use LibEnsemble on top of WarpX
simulations. It is the entry point script that runs LibEnsemble. Libensemble
then launches WarpX simulations.

Execute locally via the following command:
    python run_libensemble_on_warpx.py --comms local --nworkers 3
On summit, use the submission script:
    bsub summit_submit_mproc.sh

The number of concurrent evaluations of the objective function will be
nworkers=1 as one worker is for the persistent gen_f.
"""

# Either 'random' or 'aposmm'
generator_type = 'aposmm'
# Either 'local' or 'summit'
machine = 'local'

import sys
import numpy as np
from warpx_simf import run_warpx  # Sim function from current directory

# Import libEnsemble modules
from libensemble.libE import libE
if generator_type == 'random':
    from libensemble.gen_funcs.sampling \
        import uniform_random_sample as gen_f
    from libensemble.alloc_funcs.give_sim_work_first\
        import give_sim_work_first as alloc_f
elif generator_type == 'aposmm':
    import libensemble.gen_funcs
    libensemble.gen_funcs.rc.aposmm_optimizers = 'nlopt'
    from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f
    from libensemble.alloc_funcs.persistent_aposmm_alloc \
        import persistent_aposmm_alloc as alloc_f
else:
    print("you shouldn' hit that")
    sys.exit()

from libensemble.tools import parse_args, save_libE_output, \
    add_unique_random_streams
from libensemble import logger
from libensemble.executors.mpi_executor import MPIExecutor

import all_machine_specs

# Import machine-specific run parameters
if machine == 'local':
    machine_specs = all_machine_specs.local_specs
elif machine == 'summit':
    machine_specs = all_machine_specs.summit_specs
else:
    print("you shouldn' hit that")
    sys.exit()

logger.set_level('INFO')

nworkers, is_manager, libE_specs, _ = parse_args()

# Set to full path of warp executable
sim_app = machine_specs['sim_app']

# Problem dimension. This is the number of input parameters exposed,
# that LibEnsemble will vary in order to minimize a single output parameter.
n = 4

exctr = MPIExecutor()
exctr.register_app(full_path=sim_app, app_name='warpx')

# State the objective function, its arguments, output, and necessary parameters
# (and their sizes). Here, the 'user' field is for the user's (in this case,
# the simulation) convenience. Feel free to use it to pass number of nodes,
# number of ranks per note, time limit per simulation etc.
sim_specs = {
    # Function whose output is being minimized. The parallel WarpX run is
    # launched from run_WarpX.
    'sim_f': run_warpx,
    # Name of input for sim_f, that LibEnsemble is allowed to modify.
    # May be a 1D array.
    'in': ['x'],
    'out': [
        # f is the single float output that LibEnsemble minimizes.
        ('f', float),
        # All parameters below are not used for calculation,
        # just output for convenience.
        # Final relative energy spread.
        ('energy_std', float, (1,)),
        # Final average energy, in MeV.
        ('energy_avg', float, (1,)),
        # Final beam charge.
        ('charge', float, (1,)),
        # Final beam emittance.
        ('emittance', float, (1,)),
        # input parameter: length of first downramp.
        ('ramp_down_1', float, (1,)),
        # input parameter: Length of second downramp.
        ('ramp_down_2', float, (1,)),
        # input parameter: position of the focusing lens.
        ('zlens_1', float, (1,)),
        # Relative strength of the lens (1. is from
        # back-of-the-envelope calculation)
        ('adjust_factor', float, (1,)),
    ],
    'user': {
        # name of input file
        'input_filename': 'inputs',
        # Run timeouts after 3 mins
        'sim_kill_minutes': 3,
        # machine-specific parameters
        'machine_specs': machine_specs
    }
}

# State the generating function, its arguments, output,
# and necessary parameters.
if generator_type == 'random':
    # Here, the 'user' field is for the user's (in this case,
    # the RNG) convenience.
    gen_specs = {
        # Generator function. Will randomly generate new sim inputs 'x'.
        'gen_f': gen_f,
        # Generator input. This is a RNG, no need for inputs.
        'in': [],
        'out': [
            # parameters to input into the simulation.
            ('x', float, (n,))
        ],
        'user': {
            # Total max number of sims running concurrently.
            'gen_batch_size': nworkers,
            # Lower bound for the n parameters.
            'lb': np.array([2.e-3, 2.e-3, 0.005, .1]),
            # Upper bound for the n parameters.
            'ub': np.array([2.e-2, 2.e-2, 0.028, 3.]),
        }
    }

    alloc_specs = {
        # Allocator function, decides what a worker should do.
        # We use a LibEnsemble allocator.
        'alloc_f': alloc_f,
        'user': {
            # If true wait for all sims to process before generate more
            'batch_mode': True,
            # Only one active generator at a time
            'num_active_gens': 1
        }
    }

elif generator_type == 'aposmm':
    # Here, the 'user' field is for the user's (in this case,
    # the optimizer) convenience.
    gen_specs = {
        # Generator function. Will randomly generate new sim inputs 'x'.
        'gen_f': gen_f,
        'persis_in': ['f', 'x', 'x_on_cube', 'sim_id', 'local_min', 'local_pt'],
        'out': [
            # parameters to input into the simulation.
            ('x', float, (n,)),
            # x scaled to a unique cube.
            ('x_on_cube', float, (n,)),
            # unique ID of simulation.
            ('sim_id', int),
            # Whether this point is a local minimum.
            ('local_min', bool),
            # whether the point is from a local optimization run
            # or a random sample point.
            ('local_pt', bool)
        ],
        'user': {
            # Number of sims for initial random sampling.
            # Optimizer starts afterwards.
            'initial_sample_size': max(nworkers-1, 1),
            # APOSMM/NLOPT optimization method
            'localopt_method': 'LN_BOBYQA',
            'num_pts_first_pass': nworkers,
            # Relative tolerance of inputs
            'xtol_rel': 1e-3,
            # Absolute tolerance of output 'f'. Determines when
            # local optimization stops.
            'ftol_abs': 3e-8,
            # Lower bound for the n input parameters.
            'lb': np.array([2.e-3, 2.e-3, 0.005, .1]),
            # Upper bound for the n input parameters.
            'ub': np.array([2.e-2, 2.e-2, 0.028, 3.]),
        }
    }

    alloc_specs = {'alloc_f': alloc_f}

else:
    print("you shouldn' hit that")
    sys.exit()

# Save H to file every N simulation evaluations
libE_specs['save_every_k_sims'] = 100
# Sim directory to be copied for each worker
libE_specs['sim_input_dir'] = 'sim'
libE_specs['sim_dirs_make'] = True
libE_specs['dedicated_mode'] = True

sim_max = machine_specs['sim_max']  # Maximum number of simulations
exit_criteria = {'sim_max': sim_max}  # Exit after running sim_max simulations

# Create a different random number stream for each worker and the manager
persis_info = add_unique_random_streams({}, nworkers + 1)

# Run LibEnsemble, and store results in history array H
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria,
                            persis_info, alloc_specs, libE_specs)

# Save results to numpy file
if is_manager:
    save_libE_output(H, persis_info, __file__, nworkers)



import os
import re
import shutil
from itertools import groupby
from operator import itemgetter

from libensemble.utils.loc_stack import LocationStack
from libensemble.tools.fields_keys import libE_spec_sim_dir_keys, libE_spec_gen_dir_keys, \
    libE_spec_calc_dir_misc
from libensemble.message_numbers import EVAL_SIM_TAG, calc_type_strings


class EnsembleDirectory:
    """
    The EnsembleDirectory class provides methods for workers to initialize and
    manipulate the optional output directories where workers can change to before
    calling user functions.

    The top-level ensemble directory typically stores unique sub-directories containing results
    for each libEnsemble user function call. This can be a separate location
    on other filesystems or directories (like scratch spaces).

    When libEnsemble is initialized in a Distributed fashion, each worker can
    initiate its own ensemble directory on the local node, and copy
    back its results on completion or exception into the directory that libEnsemble
    was originally launched from.

    Ensemble directory behavior can be configured via separate libE_specs
    dictionary entries or defining an EnsembleDirectory object within libE_specs.

    Parameters
    ----------
    libE_specs: dict
        Parameters/information for libE operations. EnsembleDirectory only extracts
        values specific for ensemble directory operations. Can technically contain
        a different set of settings then the libE_specs passed to libE().

    loc_stack: object
        A LocationStack object from libEnsemble's internal libensemble.utils.loc_stack module.
    """

    def __init__(self, libE_specs=None, loc_stack=None):

        self.specs = libE_specs
        self.loc_stack = loc_stack

        if self.specs is not None:
            self.prefix = self.specs.get('ensemble_dir_path', './ensemble')
            self.use_worker_dirs = self.specs.get('use_worker_dirs', False)
            self.sim_input_dir = self.specs.get('sim_input_dir', '').rstrip('/')
            self.sim_dirs_make = self.specs.get('sim_dirs_make', False)
            self.sim_dir_copy_files = self.specs.get('sim_dir_copy_files', [])
            self.sim_dir_symlink_files = self.specs.get('sim_dir_symlink_files', [])
            self.gen_input_dir = self.specs.get('gen_input_dir', '').rstrip('/')
            self.gen_dirs_make = self.specs.get('gen_dirs_make', False)
            self.gen_dir_copy_files = self.specs.get('gen_dir_copy_files', [])
            self.gen_dir_symlink_files = self.specs.get('gen_dir_symlink_files', [])
            self.ensemble_copy_back = self.specs.get('ensemble_copy_back', False)
            self.sim_use = any([i in self.specs for i in libE_spec_sim_dir_keys + libE_spec_calc_dir_misc])
            self.gen_use = any([i in self.specs for i in libE_spec_gen_dir_keys + libE_spec_calc_dir_misc])

    def _make_copyback_dir(self):
        """Make copyback directory, adding suffix if identical to ensemble dir"""
        copybackdir = os.path.basename(self.prefix)  # Current directory, same basename
        if os.path.relpath(self.prefix) == os.path.relpath(copybackdir):
            copybackdir += '_back'
        os.makedirs(copybackdir)

    def make_copyback_check(self):
        """Check for existing copyback, make copyback if doesn't exist"""
        try:
            os.rmdir(self.prefix)
        except FileNotFoundError:
            pass
        except Exception:
            raise
        if self.ensemble_copy_back:
            self._make_copyback_dir()

    def use_calc_dirs(self, type):
        """Determines calc_dirs enabling for each calc type"""

        if type == EVAL_SIM_TAG:
            return self.sim_use
        else:
            return self.gen_use

    @staticmethod
    def extract_H_ranges(Work):
        """Convert received H_rows into ranges for labeling """
        work_H_rows = Work['libE_info']['H_rows']
        if len(work_H_rows) == 1:
            return str(work_H_rows[0])
        else:
            # From https://stackoverflow.com/a/30336492
            ranges = []
            for diff, group in groupby(enumerate(work_H_rows.tolist()), lambda x: x[0]-x[1]):
                group = list(map(itemgetter(1), group))
                if len(group) > 1:
                    ranges.append(str(group[0]) + '-' + str(group[-1]))
                else:
                    ranges.append(str(group[0]))
            return '_'.join(ranges)

    def _make_calc_dir(self, workerID, H_rows, calc_str, locs):
        """Create calc dirs and intermediate dirs, copy inputs, based on libE_specs"""

        if calc_str == 'sim':
            input_dir = self.sim_input_dir
            do_calc_dirs = self.sim_dirs_make
            copy_files = self.sim_dir_copy_files
            symlink_files = self.sim_dir_symlink_files
        else:  # calc_str is 'gen'
            input_dir = self.gen_input_dir
            do_calc_dirs = self.gen_dirs_make
            copy_files = self.gen_dir_copy_files
            symlink_files = self.gen_dir_symlink_files

        # If 'use_worker_dirs' only calc_dir option. Use worker dirs, but no calc dirs
        if self.use_worker_dirs and not self.sim_dirs_make and not self.gen_dirs_make:
            do_calc_dirs = False

        # If using input_dir, set of files to copy is contents of provided dir
        if input_dir:
            copy_files = set(copy_files + [os.path.join(input_dir, i) for i in os.listdir(input_dir)])

        # If identical paths to copy and symlink, remove those paths from symlink_files
        if len(symlink_files):
            symlink_files = [i for i in symlink_files if i not in copy_files]

        # Cases where individual sim_dirs or gen_dirs not created.
        if not do_calc_dirs:
            if self.use_worker_dirs:  # Each worker does work in worker dirs
                key = workerID
                dir = "worker" + str(workerID)
                prefix = self.prefix
            else:  # Each worker does work in prefix (ensemble_dir)
                key = self.prefix
                dir = self.prefix
                prefix = None

            locs.register_loc(key, dir, prefix=prefix, copy_files=copy_files,
                              symlink_files=symlink_files, ignore_FileExists=True)
            return key

        # All cases now should involve sim_dirs or gen_dirs
        # ensemble_dir/worker_dir registered here, set as parent dir for calc dirs
        if self.use_worker_dirs:
            worker_dir = "worker" + str(workerID)
            worker_path = os.path.abspath(os.path.join(self.prefix, worker_dir))
            calc_dir = calc_str + str(H_rows)
            locs.register_loc(workerID, worker_dir, prefix=self.prefix)
            calc_prefix = worker_path

        # Otherwise, ensemble_dir set as parent dir for sim dirs
        else:
            calc_dir = "{}{}_worker{}".format(calc_str, H_rows, workerID)
            if not os.path.isdir(self.prefix):
                os.makedirs(self.prefix, exist_ok=True)
            calc_prefix = self.prefix

        # Register calc dir with adjusted parent dir and source-file location
        locs.register_loc(calc_dir, calc_dir,  # Dir name also label in loc stack dict
                          prefix=calc_prefix,
                          copy_files=copy_files,
                          symlink_files=symlink_files)

        return calc_dir

    def prep_calc_dir(self, Work, calc_iter, workerID, calc_type):
        """Determines choice for calc_dir structure, then performs calculation."""

        if not self.loc_stack:
            self.loc_stack = LocationStack()

        if calc_type == EVAL_SIM_TAG:
            H_rows = self.extract_H_ranges(Work)
        else:
            H_rows = str(calc_iter[calc_type])

        calc_str = calc_type_strings[calc_type]

        calc_dir = self._make_calc_dir(workerID, H_rows, calc_str, self.loc_stack)

        return self.loc_stack, calc_dir

    def copy_back(self):
        """Copy back all ensemble dir contents to launch location"""
        if os.path.isdir(self.prefix) and self.ensemble_copy_back:

            no_calc_dirs = not self.sim_dirs_make or \
                not self.gen_dirs_make

            copybackdir = os.path.basename(self.prefix)

            if os.path.relpath(self.prefix) == os.path.relpath(copybackdir):
                copybackdir += '_back'

            for dir in self.loc_stack.dirs.values():
                dest_path = os.path.join(copybackdir, os.path.basename(dir))
                if dir == self.prefix:  # occurs when no_calc_dirs is True
                    continue  # otherwise, entire ensemble dir copied into copyback dir

                shutil.copytree(dir, dest_path, symlinks=True)
                if os.path.basename(dir).startswith('worker'):
                    return  # Worker dir (with all contents) has been copied.

            # If not using calc dirs, likely miscellaneous files to copy back
            if no_calc_dirs:
                p = re.compile(r"((^sim)|(^gen))\d+_worker\d+")
                for file in [i for i in os.listdir(self.prefix) if not p.match(i)]:  # each non-calc_dir file
                    source_path = os.path.join(self.prefix, file)
                    dest_path = os.path.join(copybackdir, file)
                    try:
                        if os.path.isdir(source_path):
                            shutil.copytree(source_path, dest_path, symlinks=True)
                        else:
                            shutil.copy(source_path, dest_path, follow_symlinks=False)
                    except FileExistsError:
                        continue
                    except shutil.SameFileError:  # creating an identical symlink
                        continue



"""
Runs libEnsemble with uniform random sampling and writes results into sim
dirs. This tests per-calculation sim_dir capabilities

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_sim_dirs_per_calc.py
   python3 test_sim_dirs_per_calc.py --nworkers 3 --comms local
   python3 test_sim_dirs_per_calc.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np
import os

from libensemble.libE import libE
from libensemble.tests.regression_tests.support import write_sim_func as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

sim_input_dir = './sim_input_dir'
dir_to_copy = sim_input_dir + '/copy_this'
dir_to_symlink = sim_input_dir + '/symlink_this'
c_ensemble = './ensemble_calcdirs_w' + str(nworkers) + '_' + libE_specs.get('comms')
print('creating ensemble dir: ', c_ensemble, flush=True)

for dir in [sim_input_dir, dir_to_copy, dir_to_symlink]:
    if not os.path.isdir(dir):
        os.makedirs(dir, exist_ok=True)

libE_specs['sim_dirs_make'] = True
libE_specs['ensemble_dir_path'] = c_ensemble
libE_specs['use_worker_dirs'] = False
libE_specs['sim_dir_copy_files'] = [dir_to_copy]
libE_specs['sim_dir_symlink_files'] = [dir_to_symlink]
libE_specs['ensemble_copy_back'] = True

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'out': [('x', float, (1,))],
    'user': {
        'gen_batch_size': 20,
        'lb': np.array([-3]),
        'ub': np.array([3]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 21}

H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    assert os.path.isdir(c_ensemble), 'Ensemble directory {} not created.'.format(c_ensemble)
    dir_sum = sum(['sim' in i for i in os.listdir(c_ensemble)])
    assert dir_sum == exit_criteria['sim_max'], 'Number of sim directories ({}) does not match sim_max ({}).'.format(
        dir_sum, exit_criteria['sim_max']
    )

    input_copied = []

    for base, files, _ in os.walk(c_ensemble):
        basedir = base.split('/')[-1]
        if basedir.startswith('sim'):
            input_copied.append(
                all(
                    [
                        os.path.basename(j) in files
                        for j in libE_specs['sim_dir_copy_files'] + libE_specs['sim_dir_symlink_files']
                    ]
                )
            )

    assert all(input_copied), 'Exact input files not copied or symlinked to each calculation directory'



"""
@Author: Created by Caleb Ju at Argonne National Labs as Given's associate,
         Summer 2021.
"""

import numpy as np
import scipy.sparse as spp
from libensemble.message_numbers import EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def start_consensus_persistent_gens(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    Many distributed optimization algorithms require two non-local, (e.g., not
    elementwise addition or multiplication) operations: evaluate gradients and
    consensus, which is just a linear combination of a nodes's neighbors'
    solutions. Thus, we develop a general alloc function that does both of
    these.

    From the caller function, the user must pass a square matrix in
    `persis_info['A']. Typically, this is a Laplacian matrix of a connected
    graph or a doubly stochastic matrix. Note that the alloc will send the
    row i's non-zero indices values to gen i when initializing the gen
    via the persis_info object. This information is used methods such as
    `gen_funcs/persistent_n_agent.py`, where each gen needs to take
    a linear combination of its neighbors's `x`'s.

    The user has three remaining optional values to pass into the alloc,
    which are all set from the calling script.

    - 1. The user can set `persis_info['print_progress']=1`
    This tells alloc function print the consensus value and iteration count
    whenever a consensus step is taken

    - 2. The user can set `persis_info['gen_params']={... dictionary ...}`.
    The dictionary contains the any parameters that the gen will utilize.
    This can include, for example, functions that the gen can use to compute
    gradients rather than requesting a sim to complete the work (see the
    linear regression tests in `tests/regression_tests/test_persistent_pds.py`)

    - 3. The user can set `persis_info['sim_params']={... dictionary ...}`.
    Similarly, the dictionary contains the any parameters that the sim will
    utilize

                ------------------------------------------------

    Now, we briefly explain how to request the gradient of the consensus term
    or the gradient of the function as the gen.

    For the former, the gen submits a work request and must set the parameter
    `consensus_pt` to True, while making sure to pass into the `x` variable.

    If the user wants the alloc to sum all the {f_i}, the user must set both
    `consensus_pt` and `eval_pt` to True while settings the `f_i` variable. (The
    reason for having both variables set to True is to simplify the implementation
    in the alloc.)

    Finally, to request a gradient of f_i, `consensus_pt` must be set to False,
    `get_grad` needs to be set to True, `obj_component` (i.e., which f_i to
    consider) must be set, and the `x` variable must be set.

    If the user wants a function evaluation, then set `get_grad` to False instead.
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    # Initialize alloc_specs['user'] as user.
    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    gen_count = support.count_persis_gens()
    Work = {}
    is_first_iter = False

    if persis_info.get('first_call', True):
        persis_info.update({'last_H_len': 0})
        persis_info.update({'next_to_give': 0})
        persis_info.update({'first_call': False})
        is_first_iter = True

        A = persis_info.get('A')
        assert A.shape[0] == A.shape[1], 'Matrix @A is not square'

    # Exit if all persistent gens are done
    elif gen_count == 0:
        return Work, persis_info, 1

    num_gens_at_consensus = 0

    # Sort to get consistent ordering for processing workers
    avail_persis_worker_ids = np.sort(support.avail_worker_ids(persistent=EVAL_GEN_TAG))

    for wid in avail_persis_worker_ids:

        # If at consensus, wait until everyone is done
        if persis_info[wid].get('at_consensus', False):
            num_gens_at_consensus += 1
            continue

        # Gen is waiting on sim work
        elif len(persis_info[wid].get('curr_H_ids', [])):

            [l_H_id, r_H_id] = persis_info[wid].get('curr_H_ids')
            num_sims_req = r_H_id - l_H_id

            num_fin_sims = np.sum(H['returned'][l_H_id:r_H_id])

            completed_all_sims_for_gen_i = num_fin_sims == num_sims_req

            # if completed all work, send back
            if completed_all_sims_for_gen_i:

                sims_to_ret_to_gen = np.arange(l_H_id, r_H_id)

                Work[wid] = support.gen_work(wid, ['x', 'f_i', 'gradf_i'], sims_to_ret_to_gen, persis_info.get(wid),
                                             persistent=True)

                persis_info[wid].update({'curr_H_ids': []})

        # otherwise, check if gen has requested new work
        else:
            last_H_len = persis_info['last_H_len']

            # did gen sent consensus? (start @last_H_len to avoid old work)
            consensus_sim_ids = np.where(
                np.logical_and(H[last_H_len:]['consensus_pt'], H[last_H_len:]['gen_worker'] == wid))[0]

            if len(consensus_sim_ids) > 0:
                assert len(consensus_sim_ids) == 1, 'Gen should only send one ' + \
                    'point for consensus step, received {}'.format(len(consensus_sim_ids))

                # re-center (since the last_H_len has relative index 0)
                sim_id = consensus_sim_ids[0] + last_H_len

                persis_info[wid].update({'curr_H_ids': [sim_id, sim_id + 1]})
                persis_info[wid].update({'at_consensus': True})

                num_gens_at_consensus += 1

            # otherwise, gen requested new work for sim
            else:
                new_H_ids_from_gen_i = np.where(H[last_H_len:]['gen_worker'] == wid)[0]

                assert len(new_H_ids_from_gen_i) > 0, 'Gen must request new sim ' \
                    + 'work or show convergence if avail, but neither occured'

                new_H_ids_from_gen_i += last_H_len

                l_H_id = new_H_ids_from_gen_i[0]
                r_H_id = new_H_ids_from_gen_i[-1] + 1

                # (!!)
                assert len(new_H_ids_from_gen_i) == r_H_id - l_H_id, 'new gen ' + \
                    'data must be in contiguous space'

                persis_info[wid].update({'curr_H_ids': [l_H_id, r_H_id]})

    # If all gens at consensus, distribute data to all gens
    if num_gens_at_consensus == user['num_gens']:

        assert num_gens_at_consensus == len(avail_persis_worker_ids), \
            'All gens must be available, only {}/{} are though...'.format(
                len(avail_persis_worker_ids), len(num_gens_at_consensus))

        # get index in history array @H where each gen's consensus point lies
        consensus_ids_in_H = np.array([persis_info[i]['curr_H_ids'][0] for i in avail_persis_worker_ids], dtype=int)

        # Setup for printing progress
        print_progress = persis_info.get('print_progress', False)
        print_obj = H[consensus_ids_in_H[0]]['eval_pt']
        fsum = 0

        if print_progress:
            num_gens = user['num_gens']
            n = len(gen_specs['user']['lb'])
            Ax = np.empty(num_gens * n, dtype=float)
            x = np.empty(num_gens * n, dtype=float)
            # if (1st) gen asks to gather all f_i's and print their sum

        A = persis_info['A']
        for i0, wid in enumerate(avail_persis_worker_ids):

            incident_gens = A.indices[A.indptr[i0]:A.indptr[i0 + 1]]
            # remove own index
            own_idx = np.argwhere(incident_gens == i0)
            incident_gens = np.delete(incident_gens, own_idx)

            neighbor_consensus_ids_in_H = consensus_ids_in_H[incident_gens]

            if print_progress:
                # implicilty perform matmul, $(A \kron I)[x_1,...x_m]$
                x[i0 * n:(i0 + 1) * n] = H[consensus_ids_in_H[i0]]['x']

                diag_scalar = A.diagonal()[i0]
                diag_term = diag_scalar * H[consensus_ids_in_H[i0]]['x']

                offdiag_scalars = A.data[A.indptr[i0]:A.indptr[i0 + 1]]
                offdiag_scalars = np.delete(offdiag_scalars, own_idx)
                offdiag_terms = spp.diags(offdiag_scalars).dot(H[neighbor_consensus_ids_in_H]['x'])
                offdiag_term = np.sum(offdiag_terms, axis=0)

                Ax[i0 * n:(i0 + 1) * n] = diag_term + offdiag_term

            if print_obj:
                fsum += H[consensus_ids_in_H[i0]]['f_i']

            Work[wid] = support.gen_work(wid, ['x', 'gen_worker'], np.atleast_1d(neighbor_consensus_ids_in_H),
                                         persis_info.get(wid), persistent=True)

            persis_info[wid].update({'curr_H_ids': []})
            persis_info[wid].update({'at_consensus': False})

        if print_obj and print_progress:
            msg = 'F(x)={:.8f}\n'.format(fsum)
            print('{}con={:.4e}'.format(msg, np.dot(x, Ax)), flush=True)
        elif print_obj:
            print('F(x)={:.8f}'.format(fsum), flush=True)
        elif print_progress:
            print('con={:.4e}'.format(np.dot(x, Ax)), flush=True)

    # partition sum of convex functions evenly (only do at beginning)
    if is_first_iter and len(support.avail_worker_ids(persistent=False)):
        num_funcs = user['m']
        num_gens = user['num_gens']
        num_funcs_arr = partition_funcs_arr(num_funcs, num_gens)

    inactive_workers = np.sort(support.avail_worker_ids(persistent=False))
    for i0, wid in enumerate(inactive_workers):

        # start up gens
        if is_first_iter and gen_count < user['num_gens']:

            # Checking resources first before call to gen_work
            rset_team = None
            if support.manage_resources:
                gen_resources = support.persis_info.get('gen_resources', 0)
                try:
                    rset_team = support.assign_resources(gen_resources)
                except InsufficientFreeResources:
                    break

            A = persis_info['A']

            gen_count += 1
            l_idx = num_funcs_arr[gen_count - 1]
            r_idx = num_funcs_arr[gen_count]

            A_i_indices = A.indices[A.indptr[i0]:A.indptr[i0 + 1]]
            A_i_gen_ids = inactive_workers[A_i_indices]
            # gen A_i_gen_ids[wid] corresponds to weight S_i_data[wid]
            A_i_data = A.data[A.indptr[i0]:A.indptr[i0 + 1]]

            persis_info[wid].update({
                'f_i_idxs': range(l_idx, r_idx),
                'A_i_gen_ids': A_i_gen_ids,
                'A_i_data': A_i_data,
                'params': persis_info.get('gen_params', {})})
            persis_info[wid].update({'at_consensus': False, 'curr_H_ids': []})

            Work[wid] = support.gen_work(wid, gen_specs.get('in', []), range(len(H)), persis_info.get(wid),
                                         persistent=True, rset_team=rset_team)

        # give sim work when task available
        elif persis_info['next_to_give'] < len(H):

            # skip points that are not sim work or are already done
            while persis_info['next_to_give'] < len(H) and \
                    (H[persis_info['next_to_give']]['given'] or
                     H[persis_info['next_to_give']]['consensus_pt'] or
                     H[persis_info['next_to_give']]['cancel_requested']):

                persis_info['next_to_give'] += 1

            if persis_info['next_to_give'] >= len(H):
                break

            # Checking resources first before call to sim_work
            rset_team = None
            if support.manage_resources:
                num_rsets_req = (np.max(H[persis_info['next_to_give']]['resource_sets']))
                try:
                    rset_team = support.assign_resources(num_rsets_req)
                except InsufficientFreeResources:
                    break

            gen_id = H[persis_info['next_to_give']]['gen_worker']
            [l_H_ids, r_H_ids] = persis_info[gen_id]['curr_H_ids']

            assert l_H_ids == persis_info['next_to_give'], \
                "@next_to_give={} does not match gen's requested work H id of {}".format(
                    persis_info['next_to_give'], l_H_ids)

            persis_info[wid].update({'params': persis_info.get('sim_params', {})})

            Work[wid] = support.sim_work(wid, H, sim_specs['in'], np.arange(l_H_ids, r_H_ids), persis_info.get(wid),
                                         rset_team=rset_team)

            # we can safely assume the rows are contiguous due to (!!)
            persis_info['next_to_give'] += (r_H_ids - l_H_ids)

        else:
            break

    persis_info.update({'last_H_len': len(H)})

    return Work, persis_info, 0


def partition_funcs_arr(num_funcs, num_gens):
    """ This evenly divides the functions amongst the gens. For instance,
        when there are say 7 functions and 4 gens, this function will
        distribute 2 contiguous functions to the first 3 gens, and then
        function to the last gen.

    Parameters
    ----------
    - num_funcs : int
        How {f_i}'s there are
    - num_gens : int
        How many gens

    Returns
    -------
    - num_funcs_arr : np.ndarray
        Index pointer (i.e. gen i has functions [arr[i],arr[i+1])
    """
    num_funcs_arr = (num_funcs // num_gens) * np.ones(num_gens, dtype=int)
    num_leftover_funcs = num_funcs % num_gens
    num_funcs_arr[:num_leftover_funcs] += 1

    # builds starting and ending function indices for each gen e.g. if 7
    # functions split up amongst 3 gens, then num_funcs__arr = [0, 3, 5, 7]
    num_funcs_arr = np.append(0, np.cumsum(num_funcs_arr))

    return num_funcs_arr



import sys
import os
import argparse

from libensemble.tools.tools import logger

# ==================== Command-line argument parsing ===========================

parser = argparse.ArgumentParser(prog='test_...')

parser.add_argument('--comms', type=str, nargs='?',
                    choices=['local', 'tcp', 'ssh', 'client', 'mpi'],
                    default='mpi', help='Type of communicator')
parser.add_argument('--nworkers', type=int, nargs='?',
                    help='Number of local forked processes')
parser.add_argument('--nsim_workers', type=int, nargs='?',
                    help='Number of workers for sims. 1+ zero-resource gen worker will be added')
parser.add_argument('--nresource_sets', type=int, nargs='?',
                    help='Number of resource sets')
parser.add_argument('--workers', type=str, nargs='+',
                    help='List of worker nodes')
parser.add_argument('--workerID', type=int, nargs='?', help='Client worker ID')
parser.add_argument('--server', type=str, nargs=3,
                    help='Triple of (ip, port, authkey) used to reach manager')
parser.add_argument('--pwd', type=str, nargs='?',
                    help='Working directory to be used')
parser.add_argument('--worker_pwd', type=str, nargs='?',
                    help='Working directory on remote client')
parser.add_argument('--worker_python', type=str, nargs='?',
                    default=sys.executable,
                    help='Python version on remote client')
parser.add_argument('--tester_args', type=str, nargs='*',
                    help='Additional arguments for use by specific testers')


def _get_zrw(nworkers, nsim_workers):
    """Determine zero resource workers from workers and sim workers"""
    ngen_workers = nworkers - nsim_workers
    assert ngen_workers > 0, "nsim_workers cannot be greater than number of workers"
    return [i for i in range(1, ngen_workers + 1)]


def _mpi_parse_args(args):
    """Parses arguments for MPI comms."""
    from mpi4py import MPI
    nworkers = MPI.COMM_WORLD.Get_size()-1
    is_manager = MPI.COMM_WORLD.Get_rank() == 0
    libE_specs = {'mpi_comm': MPI.COMM_WORLD, 'comms': 'mpi'}

    if args.nresource_sets is not None:
        libE_specs['num_resource_sets'] = args.nresource_sets

    # Convenience option which sets other libE_specs options.
    nsim_workers = args.nsim_workers
    if nsim_workers is not None:
        libE_specs['zero_resource_workers'] = _get_zrw(nworkers, nsim_workers)

    return nworkers, is_manager, libE_specs, args.tester_args


def _local_parse_args(args):
    """Parses arguments for forked processes using multiprocessing."""

    libE_specs = {'comms': 'local'}
    nworkers = args.nworkers

    if args.nresource_sets is not None:
        libE_specs['num_resource_sets'] = args.nresource_sets

    # Convenience option which sets other libE_specs options.
    nsim_workers = args.nsim_workers
    if nsim_workers is not None:
        nworkers = nworkers or nsim_workers + 1
        libE_specs['zero_resource_workers'] = _get_zrw(nworkers, nsim_workers)

    nworkers = nworkers or 4
    libE_specs['nworkers'] = nworkers

    return nworkers, True, libE_specs, args.tester_args


def _tcp_parse_args(args):
    """Parses arguments for local TCP connections"""
    nworkers = args.nworkers or 4
    cmd = [
        sys.executable, sys.argv[0], "--comms", "client", "--server",
        "{manager_ip}", "{manager_port}", "{authkey}", "--workerID",
        "{workerID}", "--nworkers",
        str(nworkers)]
    libE_specs = {'nworkers': nworkers, 'worker_cmd': cmd, 'comms': 'tcp'}
    return nworkers, True, libE_specs, args.tester_args


def _ssh_parse_args(args):
    """Parses arguments for SSH with reverse tunnel."""
    nworkers = len(args.workers)
    worker_pwd = args.worker_pwd or os.getcwd()
    script_dir, script_name = os.path.split(sys.argv[0])
    worker_script_name = os.path.join(worker_pwd, script_name)
    ssh = [
        "ssh", "-R", "{tunnel_port}:localhost:{manager_port}", "{worker_ip}"]
    cmd = [
        args.worker_python, worker_script_name, "--comms", "client",
        "--server", "localhost", "{tunnel_port}", "{authkey}", "--workerID",
        "{workerID}", "--nworkers",
        str(nworkers)]
    cmd = " ".join(cmd)
    cmd = "( cd {} ; {} )".format(worker_pwd, cmd)
    ssh.append(cmd)
    libE_specs = {'workers': args.workers,
                  'worker_cmd': ssh,
                  'ip': 'localhost',
                  'comms': 'tcp'}
    return nworkers, True, libE_specs, args.tester_args


def _client_parse_args(args):
    """Parses arguments for a TCP client."""
    nworkers = args.nworkers or 4
    ip, port, authkey = args.server
    libE_specs = {'ip': ip,
                  'port': int(port),
                  'authkey': authkey.encode('utf-8'),
                  'workerID': args.workerID,
                  'nworkers': nworkers,
                  'comms': 'tcp'}
    return nworkers, False, libE_specs, args.tester_args


def parse_args():
    """
    Parses command-line arguments.

    .. code-block:: python

        from libensemble.tools import parse_args
        nworkers, is_manager, libE_specs, misc_args = parse_args()

    From the shell::

        $ python calling_script --comms local --nworkers 4

    Usage:

    .. code-block:: bash

        usage: test_... [-h] [--comms [{local,tcp,ssh,client,mpi}]]
                        [--nworkers [NWORKERS]] [--workers WORKERS [WORKERS ...]]
                        [--nsim_workers [NSIM_WORKERS]]
                        [--nresource_sets [NRESOURCE_SETS]]
                        [--workerID [WORKERID]] [--server SERVER SERVER SERVER]
                        [--pwd [PWD]] [--worker_pwd [WORKER_PWD]]
                        [--worker_python [WORKER_PYTHON]]
                        [--tester_args [TESTER_ARGS [TESTER_ARGS ...]]]

    Returns
    -------

    nworkers: :obj:`int`
        Number of workers libEnsemble will initiate

    is_manager: :obj:`boolean`
        Indicates whether the current process is the manager process

    libE_specs: :obj:`dict`
        Settings and specifications for libEnsemble
        :doc:`(example)<data_structures/libE_specs>`

    """
    args, unknown = parser.parse_known_args(sys.argv[1:])
    front_ends = {
        'mpi': _mpi_parse_args,
        'local': _local_parse_args,
        'tcp': _tcp_parse_args,
        'ssh': _ssh_parse_args,
        'client': _client_parse_args}
    if args.pwd is not None:
        os.chdir(args.pwd)
    nworkers, is_manager, libE_specs, tester_args = front_ends[args.comms or 'mpi'](args)
    if is_manager and unknown:
        logger.warning('parse_args ignoring unrecognized arguments: {}'.format(' '.join(unknown)))
    return nworkers, is_manager, libE_specs, tester_args



# From https://stackoverflow.com/questions/4716533/how-to-attach-debugger-to-a-python-subproccess

# Usage:
# from libensemble.tools import ForkablePdb
# ForkablePdb().set_trace()

import pdb
import sys


# Best implementation depends on system.
# If one of these does not work, try the other.
class ForkablePdb(pdb.Pdb):
    """A Pdb subclass that may be used
    from a forked multiprocessing child

    Usage:

    .. code-block:: python

        from libensemble.tools import ForkablePdb
        ForkablePdb().set_trace()

    """
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin


# class ForkablePdb(pdb.Pdb):
#     """A Pdb subclass that may be used
#     #from a forked multiprocessing child"""
#     _original_stdin_fd = sys.stdin.fileno()
#     _original_stdin = None

#     def __init__(self):
#         pdb.Pdb.__init__(self, nosigint=True)

#     def _cmdloop(self):
#         current_stdin = sys.stdin
#         try:
#             if not self._original_stdin:
#                 self._original_stdin = os.fdopen(self._original_stdin_fd)
#             sys.stdin = self._original_stdin
#             self.cmdloop()
#         finally:
#             sys.stdin = current_stdin



#!/usr/bin/env python
"""
Parallel Hello World
"""

from mpi4py import MPI
import sys
import time

size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()
name = MPI.Get_processor_name()

time.sleep(2)
sys.stdout.write("Hello, World after sleep! I am process %d of %d on %s.\n" % (rank, size, name))



import logging

from libensemble.comms.logs import LogConfig

# From https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility/35804945

MANAGER_WARNING = 35
logging.addLevelName(MANAGER_WARNING, 'MANAGER_WARNING')
logging.MANAGER_WARNING = MANAGER_WARNING


def manager_warning(self, message, *args, **kwargs):
    if self.isEnabledFor(MANAGER_WARNING):
        self._log(MANAGER_WARNING, message, args, **kwargs)


logging.Logger.manager_warning = manager_warning
LogConfig(__package__)


def set_level(level):
    """Sets libEnsemble logging level"""
    logs = LogConfig.config
    logs.set_level(level)


def get_level():
    """Returns libEnsemble logging level"""
    logs = LogConfig.config
    return logs.log_level


def set_filename(filename):
    """Sets logger filename if loggers not yet created, else None"""
    logs = LogConfig.config
    if logs.logger_set:
        logger = logging.getLogger(logs.name)
        logger.warning("Cannot set filename after loggers initialized")
    else:
        logs.filename = filename


def set_stderr_level(level):
    """ Sets logger to mirror certain messages to stderr"""
    logs = LogConfig.config
    logs.set_stderr_level(level)


def get_stderr_level():
    """ Returns libEnsemble stderr logging level """
    logs = LogConfig.config
    return logs.stderr_level



"""
Tests the 'periodic' domain use case for APOSMM with both NLopt and SciPy
local optimization methods.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_periodic.py
   python3 test_persistent_aposmm_periodic.py --nworkers 3 --comms local

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi
# TESTSUITE_NPROCS: 4
# TESTSUITE_EXTRA: true

import sys
import numpy as np

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = ['nlopt', 'scipy']
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.periodic_func import func_wrapper as sim_f
from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_out = [
    ('x', float, n),
    ('x_on_cube', float, n),
    ('sim_id', int),
    ('local_min', bool),
    ('local_pt', bool),
]

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'localopt_method': 'LN_BOBYQA',
        'xtol_abs': 1e-8,
        'ftol_abs': 1e-8,
        'lb': np.array([0, -np.pi / 2]),
        'ub': np.array([2 * np.pi, 3 * np.pi / 2]),
        'periodic': True,
        'print': True,
    },
}

alloc_specs = {'alloc_f': alloc_f}

exit_criteria = {'sim_max': 1000}

for run in range(2):
    if run == 1:
        gen_specs['user']['localopt_method'] = 'scipy_COBYLA'
        gen_specs['user']['opt_return_codes'] = [1]
        gen_specs['user'].pop('xtol_abs')
        gen_specs['user'].pop('ftol_abs')
        gen_specs['user']['scipy_kwargs'] = {'tol': 1e-8}

    persis_info = add_unique_random_streams({}, nworkers + 1)
    # Perform the run
    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

    if is_manager:
        assert persis_info[1].get('run_order'), "Run_order should have been given back"
        min_ids = np.where(H['local_min'])

        # The minima are known on this test problem. If the above [lb,ub] domain is
        # shifted/scaled to [0,1]^n, they all have value [0.25, 0.75] or [0.75, 0.25]
        minima = np.array([[0.25, 0.75], [0.75, 0.25]])
        tol = 1e-4
        for x in H['x_on_cube'][min_ids]:
            assert np.linalg.norm(x - minima[0]) < tol or np.linalg.norm(x - minima[1]) < tol



"""
Runs libEnsemble with APOSMM with the NLopt local optimizer.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_nlopt.py
   python3 test_persistent_aposmm_nlopt.py --nworkers 3 --comms local
   python3 test_persistent_aposmm_nlopt.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi tcp
# TESTSUITE_NPROCS: 3
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from math import gamma, pi, sqrt
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = 'nlopt'
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f

from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.tests.regression_tests.support import six_hump_camel_minima as minima
from time import time

nworkers, is_manager, libE_specs, _ = parse_args()

if is_manager:
    start_time = time()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_out = [
    ('x', float, n),
    ('x_on_cube', float, n),
    ('sim_id', int),
    ('local_min', bool),
    ('local_pt', bool),
]

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'sample_points': np.round(minima, 1),
        'localopt_method': 'LN_BOBYQA',
        'rk_const': 0.5 * ((gamma(1 + (n / 2)) * 5) ** (1 / n)) / sqrt(pi),
        'xtol_abs': 1e-6,
        'ftol_abs': 1e-6,
        'dist_to_bound_multiple': 0.5,
        'max_active_runs': 6,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 1000}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    print('[Manager]:', H[np.where(H['local_min'])]['x'])
    print('[Manager]: Time taken =', time() - start_time, flush=True)

    tol = 1e-5
    for m in minima:
        # The minima are known on this test problem.
        # We use their values to test APOSMM has identified all minima
        print(np.min(np.sum((H[H['local_min']]['x'] - m) ** 2, 1)), flush=True)
        assert np.min(np.sum((H[H['local_min']]['x'] - m) ** 2, 1)) < tol

    save_libE_output(H, persis_info, __file__, nworkers)



"""
The libEnsemble utilities module assists in writing consistent calling scripts
and user functions.
"""

import os
import sys
import logging
import platform
import numpy as np
import pickle

# Create logger
logger = logging.getLogger(__name__)
logger.propagate = False
logger.setLevel(logging.INFO)

# Set up format (Alt. Import LogConfig and base on that)
utils_logformat = '%(name)s: %(message)s'
formatter = logging.Formatter(utils_logformat)

# Log to file
# util_filename = 'util.log'
# fh = logging.FileHandler(util_filename, mode='w')
# fh.setFormatter(formatter)
# logger.addHandler(fh)

# Log to standard error
sth = logging.StreamHandler(stream=sys.stderr)
sth.setFormatter(formatter)
logger.addHandler(sth)

# ==================== User Sim-ID Warning =====================================

_USER_SIM_ID_WARNING = \
    ('\n' + 79*'*' + '\n' +
     "User generator script will be creating sim_id.\n" +
     "Take care to do this sequentially.\n" +
     "Information given back to the gen_f for existing sim_id values may be overwritten!\n" +
     '\n' + 79*'*' + '\n\n')

# ==================== Ensemble directory re-use error =========================

_USER_CALC_DIR_WARNING = \
    ('\n' + 79*'*' + '\n' +
     "libEnsemble attempted to reuse {} as a parent directory for calc dirs.\n" +
     "If allowed to continue, previous results may have been overwritten!\n" +
     "Resolve this by ensuring libE_specs['ensemble_dir_path'] is unique for each run." +
     '\n' + 79*'*' + '\n\n')

# ==================== Warning that persistent return data is not used ==========

_PERSIS_RETURN_WARNING = \
    ('\n' + 79*'*' + '\n' +
     "A persistent worker has returned history data on shutdown. This data is\n" +
     "not currently added to the manager's history to avoid possibly overwriting, but\n" +
     "will be added to the manager's history in a future release. If you want to\n" +
     "overwrite/append, you can set the libE_specs option ``use_persis_return_gen``\n" +
     "or ``use_persis_return_sim``"
     '\n' + 79*'*' + '\n\n')

# =================== save libE output to pickle and np ========================


def save_libE_output(H, persis_info, calling_file, nworkers, mess='Run completed'):
    """
    Writes out history array and persis_info to files.

    Format: <calling_script>_results_History_length=<length>_evals=<Completed evals>_ranks=<nworkers>

    .. code-block:: python

        save_libE_output(H, persis_info, __file__, nworkers)

    Parameters
    ----------

    H: `NumPy structured array <https://docs.scipy.org/doc/numpy/user/basics.rec.html>`_

        History array storing rows for each point.
        :doc:`(example)<data_structures/history_array>`

    persis_info: :obj:`dict`

        Persistent information dictionary
        :doc:`(example)<data_structures/persis_info>`

    calling_file  : :obj:`string`

        Name of user-calling script (or user chosen name) to prefix output files.
        The convention is to send __file__ from user calling script.

    nworkers: :obj:`int`

        The number of workers in this ensemble. Added to output file names.

    mess: :obj:`String`

        A message to print/log when saving the file.

    """

    script_name = os.path.splitext(os.path.basename(calling_file))[0]
    short_name = script_name.split("test_", 1).pop()
    prob_str = 'length=' + str(len(H)) \
                         + '_evals=' + str(sum(H['returned'])) \
                         + '_workers=' + str(nworkers)

    h_filename = short_name + '_history_' + prob_str
    p_filename = short_name + '_persis_info_' + prob_str

    status_mess = ' '.join(['------------------', mess, '-------------------'])
    logger.info('{}\nSaving results to file: {}'.format(status_mess, h_filename))
    np.save(h_filename, H)

    with open(p_filename + ".pickle", "wb") as f:
        pickle.dump(persis_info, f)

# ===================== per-process numpy random-streams =======================


def add_unique_random_streams(persis_info, nstreams, seed=''):
    """
    Creates nstreams random number streams for the libE manager and workers
    when nstreams is num_workers + 1. Stream i is initialized with seed i by default.
    Otherwise the streams can be initialized with a provided seed.

    The entries are appended to the provided persis_info dictionary.

    .. code-block:: python

        persis_info = add_unique_random_streams(old_persis_info, nworkers + 1)

    Parameters
    ----------

    persis_info: :obj:`dict`

        Persistent information dictionary
        :doc:`(example)<data_structures/persis_info>`

    nstreams: :obj:`int`

        Number of independent random number streams to produce

    seed: :obj:`int`

        (Optional) Seed for identical random number streams for each worker. If
        explicitly set to ``None``, random number streams are unique and seed
        via other pseudorandom mechanisms.

    """

    for i in range(nstreams):

        if isinstance(seed, int) or seed is None:
            random_seed = seed
        else:
            random_seed = i

        if i in persis_info:
            persis_info[i].update({
                'rand_stream': np.random.RandomState(random_seed),
                'worker_num': i})
        else:
            persis_info[i] = {
                'rand_stream': np.random.RandomState(random_seed),
                'worker_num': i}
    return persis_info


# A very specific exception to using the logger.
def eprint(*args, **kwargs):
    """Prints a user message to standard error"""
    print(*args, file=sys.stderr, **kwargs)


# ===================== OSX set multiprocessing start =======================
# On Python 3.8 on macOS, the default start method for new processes was
#  switched to 'spawn' by default due to 'fork' potentially causing crashes.
# These crashes haven't yet been observed with libE, but with 'spawn' runs,
#  warnings about leaked semaphore objects are displayed instead.
# The next several statements enforce 'fork' on macOS (Python 3.8)
def osx_set_mp_method():
    if platform.system() == 'Darwin':
        from multiprocessing import set_start_method
        set_start_method('fork', force=True)



"""
Runs libEnsemble to test basic worker/manager communications
Scale up array_size and number of workers as required

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_comms.py
   python3 test_comms.py --nworkers 3 --comms local
   python3 test_comms.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be N-1.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.comms_testing import float_x1000 as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor  # Only used to get workerID in float_x1000

nworkers, is_manager, libE_specs, _ = parse_args()
libE_specs['disable_resource_manager'] = True
exctr = MPIExecutor()

array_size = int(1e6)  # Size of large array in sim_specs
rounds = 2  # Number of work units for each worker
sim_max = nworkers * rounds

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('arr_vals', float, array_size), ('scal_val', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, (2,))],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'gen_batch_size': sim_max,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': sim_max, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    assert flag == 0
    for i in range(sim_max):
        x1 = H['x'][i][0] * 1000.0
        x2 = H['x'][i][1]
        assert np.all(H['arr_vals'][i] == x1), "Array values do not all match"
        assert H['scal_val'][i] == x2 + x2 / 1e7, "Scalar values do not all match"

    save_libE_output(H, persis_info, __file__, nworkers)



import numpy as np
import libensemble.gen_funcs.old_aposmm as al
import libensemble.tests.unit_tests.setup as setup

n = 2

gen_out = [('x', float, n),
           ('x_on_cube', float, n),
           ('sim_id', int),
           ('priority', float),
           ('local_pt', bool),
           ('known_to_aposmm', bool),
           ('dist_to_unit_bounds', float),
           ('dist_to_better_l', float),
           ('dist_to_better_s', float),
           ('ind_of_better_l', int),
           ('ind_of_better_s', int),
           ('started_run', bool),
           ('num_active_runs', int),
           ('local_min', bool)]


def test_failing_localopt_method():
    hist, sim_specs_0, gen_specs_0, exit_criteria_0, alloc = setup.hist_setup1()

    hist.H['returned'] = 1

    gen_specs_0['user']['localopt_method'] = 'BADNAME'

    try:
        al.advance_local_run(hist.H, gen_specs_0['user'], 0, 0, {'run_order': {0: [0, 1]}})
    except al.APOSMMException:
        assert 1, "Failed like it should have"
    else:
        assert 0, "Didn't fail like it should have"


def test_exception_raising():
    hist, sim_specs_0, gen_specs_0, exit_criteria_0, alloc = setup.hist_setup1(n=2)
    hist.H['returned'] = 1

    for method in ['LN_SBPLX', 'pounders', 'scipy_COBYLA']:
        gen_specs_0['user']['localopt_method'] = method

        out = al.advance_local_run(hist.H, gen_specs_0['user'], 0, 0, {'run_order': {0: [0, 1]}})

        assert out[0] == 0, "Failed like it should have"


def test_decide_where_to_start_localopt():
    H = np.zeros(10, dtype=gen_out + [('f', float), ('returned', bool)])
    H['x'] = np.random.uniform(0, 1, (10, 2))
    H['f'] = np.random.uniform(0, 1, 10)
    H['returned'] = 1

    b = al.decide_where_to_start_localopt(H, 9, 1)
    assert len(b) == 0

    b = al.decide_where_to_start_localopt(H, 9, 1, nu=0.01)
    assert len(b) == 0


def test_calc_rk():
    rk = al.calc_rk(2, 10, 1)

    rk = al.calc_rk(2, 10, 1, 10)
    assert np.isinf(rk)


def test_initialize_APOSMM():
    hist, sim_specs_0, gen_specs_0, exit_criteria_0, alloc = setup.hist_setup1()

    al.initialize_APOSMM(hist.H, gen_specs_0)


def test_declare_opt():
    hist, sim_specs_0, gen_specs_0, exit_criteria_0, alloc = setup.hist_setup1(n=2)

    try:
        al.update_history_optimal(hist.H['x_on_cube'][0]+1, hist.H, np.arange(0, 10))
    except AssertionError:
        assert 1, "Failed because the best point is not in H"
    else:
        assert 0

    hist.H['x_on_cube'][1] += np.finfo(float).eps
    hist.H['f'][1] -= np.finfo(float).eps

    # Testing case where point near x_opt is slightly better.
    al.update_history_optimal(hist.H['x_on_cube'][0], hist.H, np.arange(0, 10))
    assert np.sum(hist.H['local_min']) == 2


def test_localopt_error_saving():
    _, sim_specs_0, gen_specs_0, _, _ = setup.hist_setup1()

    H = np.zeros(4, dtype=gen_out + [('f', float), ('fvec', float, 2), ('returned', bool)])
    H['x'] = np.random.uniform(0, 1, (4, 2))
    H['f'] = np.random.uniform(0, 1, 4)
    H['returned'] = True
    H['local_pt'][1:] = True
    gen_specs_0['user']['initial_sample_size'] = 1
    gen_specs_0['user']['localopt_method'] = 'scipy_COBYLA'
    gen_specs_0['user']['tol'] = 0.1
    gen_specs_0['user']['ub'] = np.ones(2)
    gen_specs_0['user']['lb'] = np.zeros(2)

    persis_info_1 = {'run_order': {0: [1, 2, 3]},
                     'old_runs': {},
                     'total_runs': 0,
                     'rand_stream': np.random.RandomState(1)}

    try:
        al.aposmm_logic(H, persis_info_1, gen_specs_0, _)
    except Exception as e:
        assert e.args[0] == 'Exit code is 0, but x_new was not updated in local opt run 0 after 3 evaluations.\n'\
                            'Saving run information to: run_0_abort.pickle\nWorker crashing!'
    else:
        assert 0


if __name__ == "__main__":
    test_localopt_error_saving()
    print('done')
    test_failing_localopt_method()
    print('done')
    test_exception_raising()
    print('done')
    test_decide_where_to_start_localopt()
    print('done')
    test_calc_rk()
    print('done')
    test_initialize_APOSMM()
    print('done')
    test_declare_opt()
    print('done')



import libensemble.tests.unit_tests.setup as setup
from libensemble.tools.fields_keys import libE_fields
from libensemble.worker import Worker
from libensemble.message_numbers import EVAL_SIM_TAG
import numpy as np


def simfunc():
    pass


def test_worker_init_run():

    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()

    L = exit_criteria['sim_max']
    # H = np.zeros(L + len(H0), dtype=list(set(libE_fields + sim_specs['out'] + gen_specs['out'] + alloc_specs['out'])))
    H = np.zeros(L, dtype=list(set(libE_fields + sim_specs['out'] + gen_specs['out'])))

    # check
    H['sim_id'][-L:] = -1
    H['given_time'][-L:] = np.inf

    # Create work
    sim_ids = np.zeros(1, dtype=int)

    # For loop - increment sim_ids here

    Work = {'tag': EVAL_SIM_TAG, 'persis_info': {}, 'libE_info': {'H_rows': sim_ids}, 'H_fields': sim_specs['in']}
    calc_in = H[Work['H_fields']][Work['libE_info']['H_rows']]

    Worker.init_workers(sim_specs, gen_specs)

    workerID = 1

    # Testing two worker routines: init and run
    worker = Worker(workerID)
    worker.run(Work, calc_in)

    print(worker.data)


if __name__ == "__main__":
    test_worker_init_run()
